#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 13 16:43:39 2019

Utilities for processing raw DANTE data. Typical steps include:
    - attenuator correction
    - background shot subtraction
    - channel alignment (via e.g. peak finding)
    - temporal axis calibration

@author: Pawel M. Kozlowski
"""

# python modules
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import find_peaks
#from skimage import feature

# custom modules
from fiducia.misc import find_nearest
from fiducia.loader import readDanteData
import fiducia.pltDefaults


# listing all functions declared in this file so that sphinx-automodapi
# correctly documents them and doesn't document imported functions.
__all__ = ["noScope",
           "noXRD",
           "onChannels",
           "timesScope",
           "voltageScale",
           "bkgCorrect",
           "offsetCorrect",
           "attenuationFactors",
           "attenuationCorrect",
           "timeAvgBkg",
           "avgBkgCorrect",
           "polyBkg",
           "signalEdges",
           "polyBkgFrame",
           "highestPeak",
           "highestN",
           "getPeaks",
           "alignPeaks",
           "constructMeasurementFrame",
           "loadCorrected",
           "hysteresisCorrect",
           "align",
           ]


def noScope(hf):
    r"""
    Given a header frame, return a list of channels with no scope. These
    are the channels that are off.
    
    Parameters
    ----------
    hf: pandas.core.frame.DataFrame
        Header frame from DANTE measurement data. See readDanteData().
    
    Returns
    -------
    set
        Set of channels corresponding to oscilloscopes marked as "off"
        in the Dante data file header.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    totalChannels = int(np.shape(hf)[1] / 2)
    offScope = []
    for idx in range(totalChannels):
        ch = idx + 1
        chStr = str(ch)
        if hf[chStr]['Scope type'] == 9999:
            offScope.append(ch)
    return set(offScope)


def noXRD(hf):
    r"""
    Given a header frame, return a list of channels with no XRD. If there is
    a scope, then these channels may still register a signal!
    
    Parameters
    ----------
    hf: pandas.core.frame.DataFrame
        Header frame from DANTE measurement data. See readDanteData().
    
    Returns
    -------
    set
        Set of channels corresponding to no XRDs marked in the Dante data file
        header.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    totalChannels = int(np.shape(hf)[1] / 2)
    offXRD = []
    for idx in range(totalChannels):
        ch = idx + 1
        chStr = str(ch)
        if hf[chStr]['XRD SN'] == 0:
            offXRD.append(ch)
    return set(offXRD)


def onChannels(hf):
    r"""
    Given a header frame, return a list of which dante channels
    were on for the shot.
    
    Parameters
    ----------
    hf: pandas.core.frame.DataFrame
        Header frame from DANTE measurement data. See readDanteData().
    
    Returns
    -------
    set
        Set of channels corresponding oscilloscopes and XRDs marked as on
        within the Dante data file header.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    offXRDSet = noXRD(hf)
    offScopeSet = noScope(hf)
    allOff = offXRDSet.union(offScopeSet)
    allChSet = set(np.arange(18) + 1)
    onChSet = allChSet - allOff
    return onChSet


def chStatus(hf):
    r"""
    Generates a list of statuses for each channel (on, no scope, or no XRD).
    
    Parameters
    ----------
    hf: pandas.core.frame.DataFrame
        Header frame from DANTE measurement data. See readDanteData().
    
    Returns
    -------
    status: dict
        Statuses of Dante channels according to data file header.
    
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    offXRDSet = noXRD(hf)
    offScopeSet = noScope(hf)
    allOff = offXRDSet.union(offScopeSet)
    allChArr = np.arange(18) + 1
    allChSet = set(allChArr)
    onChSet = allChSet - allOff
    status = {}
    for chNum in allChArr:
        if chNum in onChSet:
            status[chNum] = "On"
        elif chNum in offXRDSet:
            status[chNum] = "No XRD"
        elif chNum in offScopeSet:
            status[chNum] = "No Scope"
        else:
            status[chNum] = "Off?"
    return status


def timesScope(hf):
    r"""
    Given a headerFrame, returns a timesFrame containing an array of
    oscilloscope times for each channel and background shot in the
    headerFrame.
    
    Parameters
    ----------
    hf: pandas.core.frame.DataFrame
        Header frame from DANTE measurement data. See readDanteData().
    
    Returns
    -------
    timesFrame: pandas.core.frame.DataFrame
        Returns a timesFrame containing the corresponding times for each
        oscilloscope trace contained in the header frame.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    timesFrame = pd.DataFrame(index=np.arange(1024), columns=hf.keys())
    for key in hf.keys():
        timeCode = str(hf[key]['Full scale Hor time'])
        # if the time code is 99, it means the scope was off
        if timeCode == '99':
            print(f"No timespan for channel {key}.")
        # extracting total time spanned by the oscilloscope trace based on the
        # time code. For example, a time code of 208 means 2.0e-8 seconds.
        else:
            # assuming decimal point comes after first number, and all but the last
            # number are part of the significand.
            timeSpan = float(timeCode[0] + '.' + timeCode[1:-1] + 'e-' + timeCode[-1])
            # getting total number of points recorded by oscilloscope
            scopePts = hf[key]['#Hor Pts']
            # getting length of time steps between points by dividing the total time
            # over which oscilloscope trace is recorded by the total number of
            # time steps. There is one fewer step than number of points.
            timeStep = timeSpan / (scopePts - 1)
            # constructing array of times for convience in analysis/plotting
            timesFrame[key] = np.arange(scopePts) * timeStep

    return timesFrame


def voltageScale(hf, df):
    r"""
    Scales voltage (vertical) axis of dante signals based on information
    contained in the header. Returns a dataframe with the dante signals
    in units of volts. Also returns an errors/uncertainties frame
    in units of volts, where the uncertainty due to the 11-bit ADC
    converter has been calculated.
    
    
    Parameters
    ----------
    hf: pandas.core.frame.DataFrame
        Header dataframe from dante .dat file. See readDanteData().
        
    df: pandas.core.frame.DataFrame
        Dante dataframe. See readDanteData().
    
    Returns
    -------
    dfScaled: pandas.core.frame.DataFrame
        Dante dataframe with signals in units of volts.
        
    errFrame: pandas.core.frame.DataFrame
        Correspoding errors for dfScaled. Also in units of volts.
    
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    # number of bits in ADC converter
    bits = 11
    # number of bins/segments that voltage range is divided into. This
    # defines the uncertainty in the measurements due to the oscilloscope
    # and is one less than the total number of points spanning the range
    # of the oscilloscope's ADC.
    bitRange = 2 ** bits - 1
    # initializing scaled dataframe and error frame with same structure
    # as input dataframe.
    dfScaled = pd.DataFrame().reindex_like(df)
    errFrame = pd.DataFrame().reindex_like(df)
    for key in hf.keys():
        # channel voltage scale in units of volts
        maxVoltage = hf[key]['Full Scale Vert mV'] / 1e3
        # scale conversion factor
        countsToVolts = maxVoltage / bitRange
        # converting oscilloscope traces from counts to volts
        dfScaled[key] = df[key] * countsToVolts
        # uncertainties are constant and based on the ADC resolution
        # for each channel
        errFrame[key] = np.ones_like(df[key]) * countsToVolts
    return dfScaled, errFrame


def bkgCorrect(df, timesFrame):
    r"""
    Give a Dante data frame containing measurement data and background
    shot data, remove the background from the data and return the corrected
    data as a dataframe. Note that the returned dataframe is different
    in a few ways from the input dataframe. First, the returned dataframe
    is assumed to have strings as column headers, whereas the returned
    dataframe will have integers (corresponding to dante channel
    number) as the column headers. In addition, the input dataframe
    will start indexing at some number above 0 (usually 18, due to the
    header length), whereas the returned dataframe is re-indexed to
    begin at 0.
    
    A dataframe with corresponding time scales to df is also passed to this
    function for reindexing from strings to integers. This also acts as a
    placeholder in case it is necessary to interpolate values if the
    background shot and measurement shot timescales are not the same. Though
    this type of interpolation is not currently implemented.
    
    
    Parameters
    ----------
    df: pandas.core.frame.DataFrame
        Dataframe of raw dante data. This should contain both the shot
        measurement and the shot background as columns. See readDanteData().
        The columns in this dataframe are assumed to be strings.
        
    timesFrame: pandas.core.frame.DataFrame
        Dataframe containing time axis corresponding to dante signals in
        df dataframe. See timesScope().
    
    Returns
    -------
    timesBkg: pandas.core.frame.DataFrame
        Returns a dataframe of times corresponding to dfCorrected signals.
        The columns in this dataframe are integers corresponding to Dante
        channel number.
    
    dfCorrected: pandas.core.frame.DataFrame
        Returns a dataframe of background subtracted dante signals. The
        columns in this dataframe are integers corresponding to Dante
        channel number.
    
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    # reassign index (dataframe currently starts at 18 or so due to header)
    dfnew = df.set_index(np.arange(np.shape(df)[0]))
    # get total number of channels in dataframe
    totalChannels = int(np.shape(dfnew)[1] / 2)
    # produce a list of all channels
    channels = np.arange(totalChannels) + 1
    # get the total number of datapoints in the dataframe
    dataRange = np.arange(np.shape(dfnew)[0])
    # initialize an empty dataframe for storing background corrected values.
    dfCorrected = pd.DataFrame(index=dataRange, columns=channels)
    # initializing a new timesframe
    timesBkg = pd.DataFrame(index=dataRange, columns=channels)
    for ch in channels:
        # subtract shot background from shot measurement.
        dfCorrected[ch] = dfnew[str(ch)] - dfnew[str(ch) + ' bkg']
        # passing through time values to new frame
        timesBkg[ch] = timesFrame[str(ch)]
    return timesBkg, dfCorrected


def offsetCorrect(df, timesFrame, offsetsFile):
    r"""
    Reads given offset correction file (.xls) and applies offsets to
    dante measurement data given in dataframe. The input dataframe
    should already be background corrected and scaled to units of volts, 
    see bkgCorrect() and voltageScale().
    Note that although timing offsets are also applied, they are not as
    relevant since timing should be realigned to a fiducial peak
    anyway.
    Additional attenuation is not implemented and an error will be raised if
    the offsets file contains attenuation values other than 1.
    Returns a dataframe with applied offsets.
    
    Parameters
    ----------
    df: pandas.core.frame.DataFrame
        Dante dataframe with background corrected values and scaled
        to units of volts. See readDanteData(), bkgCorrect() and
        voltageScale().
        
    timesFrame: pandas.core.frame.DataFrame
        Dataframe containing time axis corresponding to dante signals in
        df dataframe. See timesScope() and bkgCorrect().
    
    
    Returns
    -------
    offsetsFile: str
        Full path to .xls file containing dante channel offsets.
    
    dfOffset: pandas.core.frame.DataFrame
        Dante dataframe with applied offset corrections
    
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    # read excel file with offsets
    offsetsFrame = pd.read_excel(offsetsFile)
    offsets = offsetsFrame.set_index(np.arange(np.shape(offsetsFrame)[0]) + 1)
    # initialize dataframe for storing offset corrected values
    dfOffset = pd.DataFrame().reindex_like(df)
    timeOffset = pd.DataFrame().reindex_like(timesFrame)
    for key in df.keys():
        # correct for channel offset
        # Offsets are added to signal
        dfOffset[key] = df[key] + offsets['OffsetV (V)'][key]
        if offsets['additional attenuation'][key] != 1:
            raise NotImplementedError("Non-zero attenuation in offsets "
                                      "file not implemented!")
        # applying time offsets
        # need to scale by 1e-9 since timesFrame is given in seconds, whereas
        # offsets are given in nanoseconds.
        timeOffset[key] = timesFrame[key] + 1e-9 * offsets['OffsetT (ns)'][key]
    return timeOffset, dfOffset


def __getAttenuation__(attenuatorsFrame, serialNum):
    r"""
    Given dataframe with attenuator info and a serial number,
    returns the attenuation factor
    
    Parameters
    ----------
    attenuatorsFrame: 
        
        
    serialNum: 
    
        
    Returns
    -------
    factorVal: 
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    factor = attenuatorsFrame[attenuatorsFrame['#'] == serialNum]['Factor']
    factorVal = factor.values
    return factorVal


def attenuationFactors(hf, channels, attenuatorsPath):
    r"""
    Given a header frame, return the attenuation factors applied to each
    channel.
    
    Parameters
    ----------
    hf: pandas.core.frame.DataFrame
        Pandas dataframe containing dante header information. See
        readDanteData().
    
    channels: set
        Set of channels to be analyzed.
        
    attenuatorsPath: str
        Full path to excel file containing attenuator serial numbers and
        corresponding attenuation factors.
    
    Returns
    -------
    chFactors
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    # load attenuators dataframe
    attenuatorsFrame = pd.read_excel(attenuatorsPath)
    # list of pandas dataframe indices containing attenuator serial numbers.
    attenuatorIdxs = ['Attenuator 1',
                      'Attenuator 2',
                      'Attenuator 3',
                      'Attenuator 4']
    # initializing array to store attenuation factors corresponding to each
    # channel.
    chFactors = pd.DataFrame(1, index=['Attenuation Factor'], columns=channels)
    # cycling over channels
    for ch in channels:
        # getting attenuator serial numbers
        serials = hf[str(ch)][attenuatorIdxs]
        # filtering null values (no attenuator)
        realSerials = serials[serials != 0]
        # fetching attenuation factors
        factors = [__getAttenuation__(attenuatorsFrame, sn) for sn in realSerials]
        # product sum of attenuation factors and saving to pandas
        # dataframe. If the list is empty then prod will return a value
        # of 1, which is exactly what we want!
        chFactors[ch] = np.prod(factors)
    return chFactors


def attenuationCorrect(attenuatorsFile, hf, df, channels):
    r"""
    Given a Dante data frame and header frame, return a data frame with
    attenuation corrections applied to each channel.
    
    Parameters
    ----------
    attenuatorsFile: str
        Full path to .xls file containing attenuator serial numbers
        and corresponding attenuation factors. See attenuationFactors().
    
    hf: pandas.core.frame.DataFrame
        Header dataframe from dante .dat file. See readDanteData().
        
    df: pandas.core.frame.DataFrame
        Dante dataframe. This frame should already be voltage scaled,
        background corrected, and offset corrected. See readDanteData().
        
    channels: list
        List of dante channels in df to be analyzed.
    
    Returns
    -------
    dfAtten: pandas.core.frame.DataFrame
        Returns dataframe with attenuation corrected signal values for
        the given channels.
    
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    
    """
    # getting attenuation factors
    attFactors = attenuationFactors(hf, channels, attenuatorsFile)
    # initializing dataframe for attenuation corrected values
    dfAtten = pd.DataFrame(index=np.arange(len(df)),
                           columns=channels)
    for ch in channels:
        # applying attenuation factors
        dfAtten[ch] = df[ch] * attFactors[ch]['Attenuation Factor']
    return dfAtten


def timeAvgBkg(times, signals, timeStart, timeEnd):
    r"""
    Calculates time averaged background for given data.
    
    Parameters
    ----------
    times:
        
        
    signals:
        
        
    timeStart:
        
        
    timeEnd:
        
    
    Returns
    -------
    avg: 
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    # getting nearest indices to start and stop times
    idxStart, valueStart = find_nearest(times, timeStart)
    idxEnd, valueEnd = find_nearest(times, timeEnd)
    # averaging
    avg = np.mean(signals[idxStart:idxEnd])
    return avg


def avgBkgCorrect(timesFrame, df, channels, timeLength=1e-9):
    r"""
    Applies background correction to bring the signal down to zero, based
    on averaging the signal background over a section of time from earliest
    time contained in timesFrame to earliest time plus timeLength.
    
    Parameters
    ----------
    timesFrame: pandas.core.frame.DataFrame
        Dataframe of time axis values corresponding to signals in df. These
        should be in units of seconds.
        
    df: pandas.core.frame.DataFrame
        Dataframe of dante signals. These should already be attenuation
        corrected and in units of volts.
        
    channels: set
        Set of channels to be analyzed.
    
    timeLength: float
        Duration of time from initial time over which to take the average. In
        units of seconds.
        
    dfAvg: pandas.core.frame.DataFrame
        Returns a dataframe containing average background corrected
        signals.
    
    Returns
    -------
    dfAvg:
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    dfAvg = pd.DataFrame().reindex_like(df)
    for ch in channels:
        # getting earliest time in frame.
        timeStart = timesFrame[ch][0]
        # added duration to earliest time to get end time for average.
        timeEnd = timeStart + timeLength
        # calculating average background by grabbing a section of the signal
        avgBkg = timeAvgBkg(times=timesFrame[ch].values,
                            signals=df[ch].values,
                            timeStart=timeStart,
                            timeEnd=timeEnd)
        # applying correction
        dfAvg[ch] = df[ch] - avgBkg
    return dfAvg


def polyBkg(time,
            signal,
            lowerEdge,
            upperEdge,
            order=3,
            lowerLength=None,
            upperLength=None,
            plot=False):
    r"""
    Fit polynomial function to ends of the signal as an estimate of the
    background signal + hyesteresis. Default is cubic fit.
    
    Parameters
    ----------
    time: numpy.ndarray
        array of times corresponding to signal
        
    signal: numpy.ndarray
        array of signal values for a single dante channel
        
    lowerEdge: int
         Index of time array corresponding to lower edge of detected
         signal. See signalEdges().
        
    upperEdge: int
        Index of time array corresponding to upper edge of detected
        signal. See signalEdges().
        
    order: int
        Order of polynomial to be fitted to estimated background/hysteresis.
        
    lowerLength: int
        Length over which to take the polynomial background fit on the
        lower end (earlier in time) segment of the signal, with respect
        to lowerEdge. Default is None, which just takes the first point
        in the signal.
        
    upperLength: int
        Length over which to take the polynomial background fit on the
        upper end (later in time) segment of the signal, with respect
        to upperEdge. Defualt is None, which then just picks the second
        to last point in the signal.
        
    plot: bool
        Flag for plotting polynomial fitted background signal. Default
        is False.
        
    Returns
    -------
    time:
        
        
    fitSignal: 
        
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    if lowerLength == None:
        lowerMin = 0
    else:
        lowerMin = lowerEdge - lowerLength
    lowerMax = lowerEdge
    upperMin = upperEdge
    if upperLength == None:
        upperMax = -1
    else:
        upperMax = upperEdge + upperLength
    dataEdgeSignal = np.concatenate((signal[lowerMin:lowerMax], signal[upperMin:upperMax]))
    dataEdgeTime = np.concatenate((time[lowerMin:lowerMax], time[upperMin:upperMax]))
    # fit polynomial
    fit = np.polyfit(dataEdgeTime, dataEdgeSignal, order)
    polyObj = np.poly1d(fit)
    fitSignal = polyObj(time)
    
    # plotting
    if plot:
        plt.plot(time, signal, label='signal')
        plt.plot(time, fitSignal, '--', label='fit')
        plt.scatter(dataEdgeTime, dataEdgeSignal, label='bkg')
        plt.legend()
        plt.show()
    
    return time, fitSignal



def signalEdges(timesFrame,
                df,
                channels,
                sigmaMult=3,
                plot=False,
                prominence=0.1,
                width=10,
                avgMult=1):
    r"""
    Determines locations and widths of peaks above the mean of the signal
    for each dante channel. Edges of the signal containing region are then
    obtained by moving sigmaMult peak widths away from the earliest and
    latest peaks. Returns these lower and upper bound edges of the signal
    containing region as a dataframe. These edges are useful for fitting
    and removing the background/hysteresis.
    
    Parameters
    ----------
    timesFrame: pandas.core.frame.DataFrame
        A dataframe containing time axis values corresponding to signals in
        df.
    
    df: pandas.core.frame.DataFrame
        A dataframe of corrected/calibrated dante signal measurements.
    
    channels: list
        List of channels in df for which edges will be determined.
    
    sigmaMult: float
        Multiplier factor by which the lower and upper bounds of the signal
        containing region are determined. The lower bound is determined
        by sigmaMult times the width of the earliest peak away from the
        earliest peak. The upper bound is determined by sigmaMult times the
        width of the latest peak away from the latest peak. Default is 3
        for approximately 3*sigma away from each peak.
        
    plot: bool
        Flag for plotting peak locations and widths. Default is False.
        
    edgesFrame: pandas.core.frame.DataFrame
        Lower and upper bound edges of the signal containing region for
        each dante channel. The lower bound is in 0 index and the upper
        bound is in 1 index. The bounds are given in index coordinates
        and they have been rounded to the nearest point.
        
    prominence: float
        Prominence threshold for identifying peaks in scipy's find_peaks().
        
    width: int
        Width in index units for identifying peaks in scipy's find_peaks().
        
    avgMult: float
        Multiplicative factor for setting minimum intensity threshold
        for indentifying peaks in scipy's find_peaks(). This is a multiple
        of the signal average.
    
    Returns
    -------
    edgesFrame: 
        
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    edgesFrame = pd.DataFrame(index=[0,1], columns=df.keys())
    for ch in channels:
        times = timesFrame[ch]
        signal = df[ch]
        # finding peaks which are above the mean of the signal
        peaks, properties = find_peaks(signal,
                                       height=avgMult * np.mean(signal),
                                       prominence=prominence,
                                       width=width)
        # determining lower and upper bounds of signal region by taking first
        # and last peaks and shifting by distance of peak width
        width1 = properties["widths"][0]
        # width / 2 is kind of like sigma, and then that multiplied by 3 is
        # like 3 sigma away.
        lowerBnd = peaks[0] - sigmaMult * width1 / 2
        width2 = properties["widths"][-1]
        upperBnd = peaks[-1] + sigmaMult * width2 / 2
        # saving these edges to the dataframe which is to be returned
        edgesFrame[ch][0] = int(round(lowerBnd))
        edgesFrame[ch][1] = int(round(upperBnd))
        if plot:
            # plotting
            plt.plot(times, signal)
            plt.plot(times[peaks], signal[peaks], "x")
            plt.vlines(x=times[peaks],
                       ymin=signal[peaks] - properties["prominences"],
                       ymax = signal[peaks],
                       color = "C1")
            plt.hlines(y=properties["width_heights"],
                       xmin=times[properties["left_ips"]],
                       xmax=times[properties["right_ips"]],
                       color = "C1")
            plt.hlines(y=avgMult * np.mean(signal),
                       xmin=times[0],
                       xmax=times[len(signal) - 1],
                       color = "C2",
                       linestyles='--')
            plt.scatter(times[edgesFrame[ch][0]],
                        signal[edgesFrame[ch][0]],
                        color="C3")
            plt.scatter(times[edgesFrame[ch][1]],
                        signal[edgesFrame[ch][1]],
                        color="C3")
            plt.title(f"Ch {ch}")
            plt.xlabel("Time (s)")
            plt.ylabel("Signal (V)")
            plt.show()
    
    return edgesFrame


def polyBkgFrame(timesFrame,
                 df,
                 edgesFrame,
                 channels,
                 order=3,
                 plot=False):
    r"""
    
    Parameters
    ----------
    timesFrame: pandas.core.frame.DataFrame
        A dataframe containing time axis values corresponding to signals in
        df.
        
    df: pandas.core.frame.DataFrame
        A dataframe of corrected/calibrated dante signal measurements.
        
    edgesFrame: pandas.core.frame.DataFrame
        Dataframe describing edges of the signal containing region, outside
        of which should be just background. This function will fit to these
        two early time and late time background containing regions. See
        signalEdges().
        
    channels: list
        A list of channels for which to apply analyis.
        
    order: int
        Order of polynomial to be fitted to estimated background/hysteresis.
        
    plot: bool
        Flag for plotting fitted background and background subtracted
        signal. Default is False.
        
    Returns
    -------
    dfPoly: 
        
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    dfPoly = pd.DataFrame().reindex_like(df)
    for ch in channels:
        # getting polynomial fit of background/hysteresis
        fitTimes, fitSignals = polyBkg(time=timesFrame[ch],
                                       signal=df[ch],
                                       lowerEdge=edgesFrame[ch][0],
                                       upperEdge=edgesFrame[ch][1],
                                       order=order,
                                       lowerLength=None,
                                       upperLength=None,
                                       plot=plot)
        # subtracting hysteresis/background and saving to dataframe
        dfPoly[ch] = df[ch] - fitSignals
        
        if plot:
            plt.plot(timesFrame[ch], dfPoly[ch])
            plt.title(f'Channel {ch}')
            plt.xlabel('Time (s)')
            plt.ylabel('Signal (V)')
            plt.show()
    return dfPoly


def highestPeak(signal, peakIdxs):
    r"""
    Find the highest peak, and return list of peaks with the highest
    peak removed from the list.
    
    Parameters
    ----------
    signal: pandas.core.series.Series
        A data series consisting of signals from a single dante channel.
        
    peakIdxs: list
        A list of indices corresponding to peaks identified in the signal
        by using scipy's find_peaks() function.
    
    Returns
    -------
    peakHighestIdx: int
        Returns the index corresponding to the highest peak.
        
    peakIdxs2: list
        Returns a list of peak index locations with the highest peak
        removed from the list. This makes it easier for highestN() to
        apply highestPeak() iteratively to find the N highest peaks.
    
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    # search for first tallest peak
    peakSignals = signal[peakIdxs]
    peakHighestIdx = pd.Series.idxmax(peakSignals)
    # finding index of tallest peak in peakIdxs array (since this is a
    # numpy array and is indexed differently from peakSignals).
    delIdx = np.where(peakIdxs==peakHighestIdx)
    # remove first tallest peak
    peakIdxs2 = np.delete(peakIdxs, delIdx)
    return peakHighestIdx, peakIdxs2 


def highestN(signal, peakIdxs, peaksNum=2):
    r"""
    Select the N tallest peaks.
    
    Parameters
    ----------
    signal: pandas.core.series.Series
        A data series consisting of signals from a single dante channel.
        
    peakIdxs: list
        A list of indices corresponding to peaks identified in the signal
        by using scipy's find_peaks() function.
        
    peaksNum: int
        Number of peaks to grab from peakIdxs. This function will grab
        just the N tallest peaks where N=peaksNum.
        
    Returns
    -------
    highestPeaks: numpy.ndarray
        Array of indices corresponding to the highest peaks. Peaks are
        ordered from highest to lowest.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    if peaksNum > len(peakIdxs):
        raise Exception("Number of peaks to grab should be less than or "
                        "equal to total number of peaks provided!")
    # initialize input to find highest peak first time around
    peakIdxsInput = peakIdxs
    # initialize array containing highest peaks
    highestPeaks = np.zeros(peaksNum, dtype=int)
    # loop over finding the N highest peaks
    for idx in range(peaksNum):
        # find the highest peak in the given list of peaks
        peakHighestIdx, peakIdxs2 = highestPeak(signal=signal,
                                                peakIdxs=peakIdxsInput)
        # reassign input to be the list of peaks without the most
        # recently found peak.
        peakIdxsInput = peakIdxs2
        # saving highest peak
        highestPeaks[idx] = peakHighestIdx
    return highestPeaks


def getPeaks(timesFrame,
             df,
             channels,
             peaksNum=2,
             plot=False,
             prominence=0.1,
             width=10,
             avgMult=1):
    r"""
    
    Parameters
    ----------
    timesFrame: pandas.core.frame.DataFrame
        A dataframe containing time axis values corresponding to signals in
        df.
        
    df: pandas.core.frame.DataFrame
        A dataframe of corrected/calibrated dante signal measurements.
    
    channels: list
        A list of channels for which to apply analyis.
        
    peaksNum: int
        Number of peaks to grab from peakIdxs. This function will grab
        just the N tallest peaks where N=peaksNum.
    
    peaksNum: int
        Number of peaks to grab from peakIdxs. This function will grab
        just the N tallest peaks where N=peaksNum.
        
    plot: bool
        Flag for plotting identified peaks, with prominences, and widths,
        overlaid with the corresponding dante signal, and the average
        dante signal.
        
    peaksFrame: pandas.core.frame.DataFrame
        Returns a dataframe containing indices of the identified peaks sorted
        from the peak that occurs earliest in time to the latest in time.
        
    prominence: float
        Prominence threshold for identifying peaks in scipy's find_peaks().
        
    width: int
        Width in index units for identifying peaks in scipy's find_peaks().
        
    avgMult: float
        Multiplicative factor for setting minimum intensity threshold
        for indentifying peaks in scipy's find_peaks(). This is a multiple
        of the signal average.
        
    
    
    Returns
    -------
    peaksFrame:
        
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    peaksFrame = pd.DataFrame(index=np.arange(peaksNum), columns=df.keys())
    for ch in channels:
        times = timesFrame[ch]
        signal = df[ch]
        # finding peaks which are above the mean of the signal
        peaks, properties = find_peaks(signal,
                                       height=avgMult * np.mean(signal),
                                       prominence=prominence,
                                       width=width)
        # selecting the N tallest peaks
        highestPeakIdxs = highestN(signal=signal,
                                   peakIdxs=peaks,
                                   peaksNum=peaksNum)
        # ordering peaks by which comes earliest
        peaksOrdered = np.sort(highestPeakIdxs)
        # saving these peaks to the dataframe which is to be returned
        peaksFrame[ch] = peaksOrdered
        if plot:
            # plotting
            plt.plot(times, signal)
            plt.plot(times[peaks], signal[peaks], "x")
            plt.vlines(x=times[peaks],
                       ymin=signal[peaks] - properties["prominences"],
                       ymax = signal[peaks],
                       color = "C1")
            plt.hlines(y=properties["width_heights"],
                       xmin=times[properties["left_ips"]],
                       xmax=times[properties["right_ips"]],
                       color = "C1")
            plt.hlines(y=avgMult * np.mean(signal),
                       xmin=times[0],
                       xmax=times[len(signal) - 1],
                       color = "C2",
                       linestyles='--')
            plt.title(f"Ch {ch}")
            plt.xlabel("Time (s)")
            plt.ylabel("Signal (V)")
            plt.show()
    return peaksFrame


def alignPeaks(timesFrame,
               df,
               peaksFrame,
               channels,
               peakAlignIdx=0,
               referenceTime=1e-9,
               plot=False):
    r"""
    Parameters
    ----------
    timesFrame: pandas.core.frame.DataFrame
        A dataframe containing time axis values corresponding to signals in
        df.
        
    df: pandas.core.frame.DataFrame
        A dataframe of corrected/calibrated dante signal measurements.
        
    peaksFrame: pandas.core.frame.DataFrame
        Dataframe containing positions of N highest peaks and sorted
        from earliest in time to latest in time. See getPeaks().
    
    peakAlignIdx: int
        Picks which peak to align to. 0 is first peak, 1 is second peak in
        peaksFrame, etc.
    
    referenceTime: float
        Time in s to which align peaks. Default is 1e-9 s or 1 ns.
        
    plot: bool
        Flag for plotting aligned dante signals. Default is False.
        
    Returns
    -------
    timesAligned: pandas.core.frame.DataFrame
        Returns a dataframe identical in shape to timesFrame, but with
        the times for each dante channel offset such that the selected
        peaks are temporally aligned.
    
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    peakAlignIdx = 0 # 0 is first peak, 1 is 2nd peak, etc.
    referenceTime = 1e-9 # set peaks to occur at 1 ns
    timesAligned = pd.DataFrame().reindex_like(timesFrame)
    for ch in channels:
        time = timesFrame[ch]
        # of the two highest peaks, grab the one with the lowest index (earliest
        # in time)
        # align all signals to reference time, by setting first peak to be at
        # reference time.
        # Get index of peak relative to signals dataframe
        peakIdx = peaksFrame[ch][peakAlignIdx]
        # get time at which peak occurs
        peakTime = time[peakIdx]
        # calculate necessary shift in peak to bring it to the reference time
        offsetTime = peakTime - referenceTime
        # generate new time scales for shifted peaks
#        print(f'times: \n{time}')
#        print(f'offset time: {offsetTime}')
        timesAligned[ch] = time - offsetTime
    # plotting
    if plot:
        for ch in channels:
            plt.plot(timesAligned[ch], df[ch], label=ch)
        plt.xlabel('Time (s)')
        plt.ylabel('Signal (V)')
        plt.title('Aligned')
        plt.legend(frameon=False,
                   labelspacing=0.001,
                   borderaxespad=0.1)
        plt.show()
    return timesAligned


def constructMeasurementFrame(timesFrame, df, channels):
    r"""
    Takes out put timesFrame and dataFrame from rawProcess.py functions
    and generates a measurementFrame that can be passed to analyzeStreak()
    and other main.py functions.
    
    Converts units from seconds to nanoseconds.
    
    Parameters
    ----------
    timesFrame: pandas.core.frame.DataFrame
        A dataframe containing time axis values corresponding to signals in
        df.
        
    df: pandas.core.frame.DataFrame
        A dataframe of corrected/calibrated dante signal measurements.
        
    channels: list
        A list of channels for which to apply analyis.
        
    Returns
    -------
    measurementFrame: pandas.core.frame.DataFrame
        Returns a measurementFrame which can be passed to main.py functions
        such as analyzeSpectrum() and analyzeStreak().
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    # generating list of column names for measurementFrame
    colsList= []
    for ch in channels:
        timeStr = 'Time' + str(ch)
        signalStr = 'Signal' + str(ch)
        colsList.append(timeStr)
        colsList.append(signalStr)
    # initialize dataframe
    measurementFrame = pd.DataFrame(index=np.arange(len(df)), columns=colsList)
    for ch in channels:
        # writing data into measurement frame columns
        measurementFrame['Time' + str(ch)] = timesFrame[ch] * 1e9
        measurementFrame['Signal' + str(ch)] = df[ch]
    return measurementFrame


def loadCorrected(danteFile,
                  attenuatorsFile,
                  offsetsFile,
                  cut=None,
                  plot=False,
                  addCh=[]):
    r"""
    Given a dante data file, an attenuators file, and an offsets file, reads
    the file and applies background correction, attenuation correction, and
    channel offset correction. Returns the corrected data traces as a pandas
    dataframe. The row indices of this dataframe also contain the 
    correct time scaling given the oscilloscope settings, but note that
    the channels are not aligned. User must apply alignment correction
    using some measured signal as a temporal fiducial.
    
    
    Parameters
    ----------
    danteFile: str
        Full path to .dat file containing raw dante traces.
        
    attenuatorsFile: str
        Full path to the .xls file containing attenuator serial numbers and
        corresponding attenuation factors.
        
    offsetsFile: str
        Full path to .xls file containing dante channel offsets.
        
    cut: int
        Number of points to cut from leading and trailing end of each
        Dante channel trace. This is used to remove noise that occurs
        at the edges of the signal. Default is None, which means no
        cut is applied.
        
    plot: bool
        Flag for plotting data after each calibration/correction step.
        Default is False.
        
    addCh: list
        Add channels to analyze. This is used to override which channels
        are listed as on in the header of the data dante data file.
    
    Returns
    -------
    timeOffset: 
        
        
    dfAvg:
        
        
    onChList:
        
        
    hf:
        
        
    dfVolt:
        
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    # load data
    hf, df = readDanteData(danteFile)
    # get set of all dante channels with useful data
    onCh = onChannels(hf)
    print(f'Analyzing channels {onCh}')
    # converting from set to list because some functions are implemented
    # assuming a list of channels instead of a set.
    if addCh:
        onChList = list(onCh) + addCh
    else:
        onChList = list(onCh)
    
    # apply voltage scaling
    dfVolt, errVolt = voltageScale(hf, df)
    
    # time axis calibration
    timesFrame = timesScope(hf)
    
    # shot background correction
    # Note: Here is where we switch from string indices to integer indices
    # since we no longer need the bkg columns in our dataframe.
    timesBkg, dfBkg = bkgCorrect(df=dfVolt,
                                 timesFrame=timesFrame)
    
    # Applying offset correction
    timeOffset, dfOffset = offsetCorrect(df=dfBkg,
                                         timesFrame=timesBkg,
                                         offsetsFile=offsetsFile)
    
    # Applying attenuation correction
    dfAtten = attenuationCorrect(attenuatorsFile, hf, dfOffset, onChList)
    
    # applying an averaged background correction
    dfAvg = avgBkgCorrect(timesFrame=timeOffset,
                          df=dfAtten,
                          channels=onCh,
                          timeLength=1e-9)
    
    # plotting each calibration/correction step
    if plot:
        plt.plot(df)
        plt.title('raw dante data, fresh off the load')
        plt.xlabel('Time (counts)')
        plt.ylabel('Signal (counts)')
        plt.show()
        
        plt.plot(dfVolt)
        plt.title('Voltage scaled')
        plt.xlabel('Time (counts)')
        plt.ylabel('Signal (V)')
        plt.show()
        
        for ch in onChList:
            plt.plot(timesFrame[str(ch)], dfVolt[str(ch)], label=ch)
        plt.title('time and voltage scaled')
        plt.xlabel('Time (s)')
        plt.ylabel('Signal (V)')
        plt.legend(frameon=False,
                   labelspacing=0.001,
                   borderaxespad=0.1)
        plt.show()
        
        for ch in onChList:
            plt.plot(timesBkg[ch], dfBkg[ch], label=ch)
        plt.title('bkg corrected')
        plt.xlabel('Time (s)')
        plt.ylabel('Signal (V)')
        plt.legend(frameon=False,
                   labelspacing=0.001,
                   borderaxespad=0.1)
        plt.show()
        
        for ch in onChList:
            plt.plot(timeOffset[ch], dfOffset[ch], label=ch)
        plt.title('offset corrected')
        plt.xlabel('Time (s)')
        plt.ylabel('Signal (V)')
        plt.legend(frameon=False,
                   labelspacing=0.001,
                   borderaxespad=0.1)
        plt.show()
        
        for ch in onChList:
            plt.plot(timeOffset[ch], dfAtten[ch], label=ch)
        plt.title('attenuation corrected')
        plt.xlabel('Time (s)')
        plt.ylabel('Signal (V)')
        plt.legend(frameon=False,
                   labelspacing=0.001,
                   borderaxespad=0.1)
        plt.show()
        
        for ch in onChList:
            plt.plot(timeOffset[ch], dfAvg[ch], label=ch)
        plt.title('averaged background corrected')
        plt.xlabel('Time (s)')
        plt.ylabel('Signal (V)')
        plt.legend(frameon=False,
                   labelspacing=0.001,
                   borderaxespad=0.1)
        plt.show()
    
    return timeOffset, dfAvg, onChList, hf, dfVolt


def hysteresisCorrect(timesFrame,
                      df,
                      channels,
                      order=5,
                      prominence=0.2,
                      width=10,
                      avgMult=1):
    r"""
    Corrects for hysteresis by detecting edges of signal containing region
    and fitting a polynomial background to regions that do not belong to 
    signal. This background is then subtracted.
    
    
    Parameters
    ----------
    timesFrame: pandas.core.frame.DataFrame
        Time corresponding to df
        
    df: pandas.core.frame.DataFrame
        Dataframe of dante signals. See loadCorrected().
        
    channels: list
        A list of channels for which to apply analyis.
        
    order: int
        Polynomial order to be fitted to hysteresis/background.
        
    prominence: float
        Prominence threshold for identifying peaks in scipy's find_peaks().
        
    width: int
        Width in index units for identifying peaks in scipy's find_peaks().
        
    avgMult: float
        Multiplicative factor for setting minimum intensity threshold
        for indentifying peaks in scipy's find_peaks(). This is a multiple
        of the signal average.
        
        
    Returns
    -------
    dfPoly: pandas.core.frame.DataFrame
        Returns a dataframe of hysteresis corrected dante signals.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    
    """
    # getting edges of signal containing region
    edgesFrame = signalEdges(timesFrame=timesFrame,
                             df=df,
                             channels=channels,
                             plot=False,
                             prominence=prominence,
                             width=width,
                             avgMult=avgMult)
    
    # removing hysteresis and background with a polynomial fit
    dfPoly = polyBkgFrame(timesFrame=timesFrame,
                          df=df,
                          edgesFrame=edgesFrame,
                          channels=channels,
                          order=order,
                          plot=False)
    return dfPoly


def align(timesFrame,
          df,
          channels,
          peaksNum=1,
          peakAlignIdx=0,
          referenceTime=1e-9,
          prominence=0.01,
          width=10,
          avgMult=1.5):
    r"""
    Aligns dante signals based on peak finding.
    
    Parameters
    ----------
    timesFrame: pandas.core.frame.DataFrame
        Time corresponding to df
        
    df: pandas.core.frame.DataFrame
        Dataframe of dante signals. See loadCorrected().
        
    channels: list
        A list of channels for which to apply analyis.
        
    peaksNum: int
        Number of peaks to grab from peakIdxs. This function will grab
        just the N tallest peaks where N=peaksNum.
        
    peakAlignIdx: int
        Picks which peak to align to. 0 is first peak, 1 is second peak in
        peaksFrame, etc.
        
    referenceTime: float
        Time in s to which align peaks. Default is 1e-9 s or 1 ns.
        
    prominence: float
        Prominence threshold for identifying peaks in scipy's find_peaks().
        
    width: int
        Width in index units for identifying peaks in scipy's find_peaks().
        
    avgMult: float
        Multiplicative factor for setting minimum intensity threshold
        for indentifying peaks in scipy's find_peaks(). This is a multiple
        of the signal average.
        
    Returns
    -------
    timesAligned: pandas.core.frame.DataFrame
        Returns a dataframe of times corresponding to signals in df, such
        that the signals are now aligned to the given peak.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    
    """
    if peakAlignIdx > peaksNum - 1:
        raise Exception(f"Cannot align to {peakAlignIdx}th peak if there"
                        f" are only {peaksNum} peaks.")
    # finding tallest peaks in the signal
    peaksFrame = getPeaks(timesFrame=timesFrame,
                          df=df,
                          channels=channels,
                          peaksNum=peaksNum,
                          plot=True,
                          prominence=prominence,
                          width=width,
                          avgMult=avgMult)
    
    # aligning to selected peak
    timesAligned = alignPeaks(timesFrame=timesFrame,
                              df=df,
                              peaksFrame=peaksFrame,
                              channels=channels,
                              peakAlignIdx=peakAlignIdx,
                              referenceTime=referenceTime,
                              plot=True)
    return timesAligned

#-----------------------------------------
# PURGATORY FOR DEPRECATED METHOD ATTEMPTS
#-----------------------------------------

#def signalEdges(time,
#                signal,
#                sigma,
#                windowMin=0,
#                windowMax=1,
#                title='',
#                plot=False):
#    """
#    Gets the signal edges via Canny edge detection method
#    
#    time: numpy.ndarray
#        Array of times corresponding to signal.
#        
#    signal: numpy.ndarray
#        Array of signal intensities.
#        
#    sigma: float
#        Width of Gaussian filter used in Canny edge detection. This should
#        be close to the width of the features you want to detect.
#        
#    windowMin: int
#        Lower bound index of window over which to search for edges in signal.
#        Default is 0.
#        
#    windowMax: int
#        Distance in index units from end of signal to upper bound of window
#        over which to search for edges in signal. Default is 1.
#        
#    plot: bool
#        Flag for plotting signal with detected edges. Default is False.
#    """
#    timeWin = time[windowMin:-windowMax]
#    signalWin = signal[windowMin:-windowMax]
#    # stacking signals to form an "image" so that scikit-image can process it.
#    sigStack = np.vstack((signalWin, signalWin, signalWin))
#    edgeCanny = feature.canny(sigStack,
#                              sigma=sigma)
#    # getting the actual edge points instead of just the indices
#    edgeTimes = timeWin[edgeCanny[1,:]]
#    edges = signalWin[edgeCanny[1,:]]
#    
#    if plot:
#        # plotting detected edges
#        plt.plot(time, signal, label='signal')
#        plt.scatter(edgeTimes, edges, color='red', label='edges')
#        plt.legend()
#        plt.title(title)
#        plt.show()
#    
#    return edgeTimes, edges
#
#
#
#
#
#def highestTwo(signal, peakIdxs):
#    """
#    Get the two highest peaks given signal, and peak
#    indices found using a peaking finder such as find_peaks_cwt().
#    """
#    # search for first tallest peak
#    peakSignals1 = signal[peakIdxs]
#    peak1Idx = np.argmax(peakSignals1)
#    # convert back into index for signals
#    peak1SigIdx = peakIdxs[peak1Idx]
#    # remove first tallest peak
#    peakIdxs2 = np.delete(peakIdxs, peak1Idx)
#    peakSignals2 = signal[peakIdxs2]
#    # search for second tallest peak
#    peak2Idx = np.argmax(peakSignals2)
#    # convert back into index for signals
#    peak2SigIdx = peakIdxs2[peak2Idx]
#    # return positions of two tallest peaks in order of which is first
#    if peak1SigIdx < peak2SigIdx:
#        return peak1SigIdx, peak2SigIdx
#    else:
#        return peak2SigIdx, peak1SigIdx


 # -*- coding: utf-8 -*-
"""
Created on Tue Jul 21 12:11:12 2020

Common statistic operations

@author: Myles T. Brophy
"""
# python modules
import numpy as np
 
# listing all functions declared in this file so that sphinx-automodapi
# correctly documents them and doesn't document imported functions.
__all__ = ["simpsVariance",
           "trapzVariance",
           "gradientVariance",
           "dotVariance",
           "interpVariance",
           ]


def simpsVariance(yUnc, x=None, dx=1.0):
    r"""
    Propagates variance (:math:`\sigma^2`) through Simpson's rule
    numerical integration.
    
    NOTE: THIS FUNCTION IS INCOMPLETE AND HAS NOT BEEN VERIFIED
    2020-08-21 PMK.
    
    Parameters
    ----------
    yUnc: list, numpy.ndarray
        Uncertainties in the vertical axis.
        
    x: list, numpy.ndarray, optional
        Horizontal coordinates corresponding to yUnc. Default is None,
        which generates a uniformly spaced linear array of horizontal
        coordinates based on the length of yUnc and the value of dx.
        
    dx: float, optional
        Uniform spacing between horizontal coordinates corresponding to
        yUnc. Default is 1.0.
        
    Returns
    -------
    variance: float
        Variance (:math:`\sigma^2`) on value of integral from Simpson's
        rule numerical integration.
        
    Notes
    -----
    Based on a modified verison of:
        https://en.wikipedia.org/wiki/Simpson's_rule#Composite_Simpson's_rule_for_irregularly_spaced_data
    
    See also
    --------
    
    Examples
    --------
    
    """
    N = len(yUnc) - 1
    yUnc = np.asarray(yUnc)
    if x is None:
        d = np.full(N, dx)
    else:
        d = np.diff(np.asarray(x))

    variance = 0.0
    #TODO validate with hand calculations
    for i in range(1, N, 2):
        hph = d[i] + d[i - 1]
        variance += yUnc[i]**2 * ( ( d[i]**3 + d[i - 1]**3
                           + 3. * d[i] * d[i - 1] * hph )\
                     / ( 6 * d[i] * d[i - 1] ) )**2
        variance += yUnc[i - 1]**2 * ( ( 2. * d[i - 1]**3 - d[i]**3
                              + 3. * d[i] * d[i - 1]**2)\
                     / ( 6 * d[i - 1] * hph) )**2
        variance += yUnc[i + 1]**2 * ( ( 2. * d[i]**3 - d[i - 1]**3
                              + 3. * d[i - 1] * d[i]**2)\
                     / ( 6 * d[i] * hph ) )**2

    if (N + 1) % 2 == 0:
        variance += yUnc[N]**2 * ( ( 2 * d[N - 1]**2
                          + 3. * d[N - 2] * d[N - 1])\
                     / ( 6 * ( d[N - 2] + d[N - 1] )) )**2
        variance += yUnc[N - 1]**2 * ( ( d[N - 1]**2
                           + 3*d[N - 1]* d[N - 2] )\
                     / ( 6 * d[N - 2] ) )**2
        variance -= yUnc[N - 2]**2 * (d[N - 1]**3\
                     / ( 6 * d[N - 2] * ( d[N - 2] + d[N - 1] )))**2
    return variance


def trapzVariance(yUnc, x=None, dx=1.0):
    r"""
    Error propogation for Trapezoidal rule integration using uniform or non-uniform grids.
        
    Parameters
    ----------
    yUnc : list, numpy.ndarray
        The list of uncertainities, referenced as :math:`\sigma_i`.
        
    x : list, numpy.ndarray, optional
        The sampling points for which the uncertainites ''y'' were found. Must be the
        same length as ''y''. If none are provided, then the step size will be uniform 
        and set with ''dx''. The default is None.
    dx : int, float, optional
        Step size. Only applies if sampling points aren't specified
        with ''x''. The default is 1.0.

    Returns
    -------
    variance : float
        The total variance (:math:`\sigma^2`) found by propagating ''y''.
        
    Notes
    -----
    Trap rule integration with non uniform spacing takes the form
    
    .. math::
        \sum_{k=1}^N \frac{\Delta x_i}{2} \left(f(x)_{i-1} + f(y)_i \right)
    
    Propogating the uncertainties through this integration results in
    
    .. math::
        \sigma^2 = \frac{1}{4} \left(\sum_{k=1}^N \Delta x_i \sigma_{i-1}^2 + \sigma_i^2 + 2\sum_{k=1}^{N-1} \Delta x_i \Delta x_i+1 \sigma_i^2 \right)
    
    The equation is generalized and applies to uniform and non-uniform step sizes.
    
    See also
    --------
    
    Examples
    --------
    """
    N = len(yUnc)
    yUnc = np.asarray(yUnc)
    if x is None:
        d = np.full(N-1, dx)
    else:
        d = np.diff(np.asarray(x))
    variance = 0.0
    #add variance of indiviual trapezoid variances
    for i in range(1, N):
        variance += d[i-1] ** 2 * 0.25 * (yUnc[i-1] ** 2 + yUnc[i] ** 2) 
    #add covariant terms because of edge overlap
    for i in range(1, N-1):
        variance += 0.5 * d[i-1] * d[i] * yUnc[i] ** 2
    return variance


def gradientVariance(yUnc, x=None, dx=1.0):
    r"""
    Propogates uncertainty for the gradient operator of an array of a given step size.

    Parameters
    ----------
    yUnc : list, numpy.ndarray
        The list of uncertainities, referenced as :math:'\sigma_i'.
        
    x : list, numpy.ndarray, optional
        The sampling points for which the uncertainites ''y'' were found. Must be the
        same length as ''y''. If none are provided, then the step size will be uniform 
        and set with ''dx''. The default is None.
        
    dx : list, numpy.ndarray, optional
        Step size. Only applies if sampling points aren't specified
        with ''x''. The default is 1.0.

    Returns
    -------
    variance : float
        The total variance (:math:`\sigma^2`) found by propagating ''y''.

    Notes
    -----
    
    .. math::
        \operatorname{Var}(\nabla y_i) = \frac{h_{i-1}^2 \sigma_{i+1}^2 + (h_i^2 + h_{i-1}^2)^2 \sigma_i^2 - h_i^4 \sigma_{i-1}^2}{(h_i h_{i-1}(h_i + h_{i-1}))^2}

    At the boundaries
    
    .. math::
        \operatorName{Var}(\nabla y_0) = \frac{\sigma_1^2 - \sigma_0^2}{h_0^2}, \operatorName{Var}(\nabla y_{N-1}) = \frac{\sigma_{N-1}^2 - \sigma_{N-2}^2}{h_{N-2}^2} 
   
    
    See also
    --------
    
    Examples
    --------
    """
    N = len(yUnc)
    np.asarray(yUnc)
    if x is None:
        h = np.full(N-1, dx)
    else:
        h = np.diff(np.asarray(x))
    
    grad = np.zeros(N)
    #square the uncertainites, cleaner to do it here than in the loop
    yUnc = yUnc**2
    grad[0] == (yUnc[1]-yUnc[0])/(h[0]**2)
    grad[N-1] == (yUnc[N-1]-yUnc[N-2])/(h[N-2]**2)
    for i in range(1, N-1):
        firstTerm = h[i-1]**4*yUnc[i+1]
        secondTerm = (h[i]**2 - h[i-1]**2)**2 * yUnc[i]
        thirdTerm = h[i]**4*yUnc[i-1]
        denom = (h[i]*h[i-1]*(h[i]+h[i-1]))**2
        
        grad[i] = (firstTerm + secondTerm - thirdTerm) / denom  

    return grad


def dotVariance(a, b, aUncertainty=None, bUncertainty=None):
    r"""
    Propogate uncertainty for the dot product of matrix a and 1D vector b.
    
    Propogate uncertainty for the dot product of a matrix and a 1D vector.
    Assumes no covariance between `a` and `b`. Methodology is similar to :func:`numpy.dot` where:
        
    - If both `a` and `b` are 1D, the uncertainty of the inner product of vectors
      is returned.
    - If 'a' is N dimensional (Where :math: `N>=2`) and `b` is 1D, the uncertainty of the sum
      product of the last axis of a with b is returned.
    
    0-D (scalar) arrays are not supported.
    `b` arrays that have more than one axis are not supported.
    `a` and `b` must have the same shape as `aUncertainty` and `bUncertainty`,
    respectively.
    
    Parameters
    ----------
    a : numpy.ndarray, list
        Matrix or vector to dot with 'b'.
  
    b : numpy.ndarray, list
        Vector that 'a' will be dotted with. Must be the same size as 
        the last axis of a.
   
    aUncertainty : numpy.ndarray, optional
        Uncertainty of each element in 'a'. The default is None.
   
    bUncertainty : numpy.ndarray, optional
        Uncertainty of each element in 'b'. The default is None.

    Returns
    -------
    variance : float
    
    Notes
    -----
    
    .. math::
        \operatorname{Var}(A \cdot B) = \sum_{i=1}^N \operatorname{Var}(a_i b_i)
    
    Assuming covariance between independent variables
    
    .. math:: 
        \sum_{i=1}^N (a_i\sigma_{b_i})^2 + (b_i\sigma_{a_i})^2
    
    See also
    --------
    
    Examples
    --------

    """
    a = np.asarray(a)
    b = np.asarray(b)
    
    if aUncertainty is None:
        aUncertainty = np.zeros(a.shape)
    else:
        aUncertainty = np.asarray(aUncertainty)
        
    if bUncertainty is None:
        bUncertainty = np.zeros(b.shape)
    else:
        bUncertainty = np.asarray(bUncertainty)
    
    #first see if the shapes are compatible 
    #also checks that 
    try:
        if b.ndim != 1:
            raise NotImplementedError("Propogating uncertainty not implemented for b arguments that aren't 1D.")
        if b.shape != bUncertainty.shape or a.shape != aUncertainty.shape:
            raise ValueError("Array does not have the same shape as it's uncertainty array.")
        np.dot(a, b)
        np.dot(a, bUncertainty)
        np.dot(aUncertainty, b)
    except ValueError:
        raise
    else:
        #uncertainty propagation for dotting vectors
        if a.ndim == 1:
            variance = np.dot(b ** 2, aUncertainty ** 2) + np.dot(a ** 2, bUncertainty ** 2)
            return variance
        #uncertainty propagation for dotting matrix with vector
        elif a.ndim >=2:
            variance = np.zeros(a.shape[:-1])
            for i in range(a.shape[0]):
                variance[i] = dotVariance(a[i], b, aUncertainty[i], bUncertainty)
            return variance
    

def interpVariance(x, xp, fpUnc, leftVar=None, rightVar=None, period=None):
    r"""
    Propagate uncertainty for linear interpolation.
    

    Parameters
    ----------
    x : numpy.ndarray, list
        The x-coordinates at which to evaluate the interpolated values.
        
    xp : numpy.ndarray, list
        The 1D x-coordinates of the data points, must be increasing order
        
    fpUnc : numpy.ndarray, list
        The uncertainty in the y-coordinates of the data points, same length as `xp`.
        
    leftVar : float, optional
        Variance to return for `x < xp[0]`. If not given, the first `yUnc` element 
        will be used. Default is `None`
        
    rightVar : float, optional
        Variance to return for `x > xp[-1]`. If not given, the last `yUnc` element 
        will be used. Default is `None`


    Returns
    -------
    yVar : numpy.ndarray
        1D array containing the variance for each interpolated `x`.

    Notes
    -----
    Variance of interpolated point, assuming no uncertainty in `x` and `xp`,
    and no covariance between y-coordinates, is given by
    
    .. math::
        \operatorname{Var}(y) = \frac{1}{(x_1 - x_0)^2} ( (x_1 - x)^2 \sigma_{y_0}^2 + (x-x_0)^2 \sigma_{y_1}^2 )
        
    
    See also
    --------
    
    Examples
    --------
    """
    #conver to ndarray
    x = np.asarray(x)
    xp = np.asarray(xp)
    fpUnc = np.asarray(fpUnc)
    
    #make space for final answer
    yVar = np.zeros(x.shape)
    
    #find the indexs where x is between xp_i and xp_{i+1}
    indices = np.searchsorted(xp, x, side='left')
    
    for i, index in enumerate(indices):
        #if index is 0 and is not equal to the first data point
        if index == 0 and x[i] != xp[0]: 
            #set to default value  
            if leftVar:
                variance = leftVar
            else:
                variance = fpUnc[0]**2
        #if the index is after the last datapoint
        elif index >= len(xp): 
            #set to default value
            if rightVar:
                variance = rightVar
            else:
                variance = fpUnc[-1]**2
        else:
           rightIndex = index
           leftIndex = index-1
           #if we already have this value
           if xp[rightIndex] == x[i]:
               
               variance = fpUnc[rightIndex] ** 2
           elif xp[leftIndex] == x[i]:
               variance = fpUnc[leftIndex] ** 2
           else:
               #new value, actually need to do interpolation propagation
               term1 = 1 / (xp[rightIndex] - xp[leftIndex]) ** 2
               term2 = (xp[rightIndex] - x[i]) ** 2 * fpUnc[leftIndex] ** 2 + (x[i] - xp[leftIndex]) ** 2 *fpUnc[rightIndex] ** 2 
               variance = term1 * term2

        yVar[i] = variance
        
    return yVar
        
        
    


"""
Test Example for Sphinx Docs
===================================

We create a test plot to see if sphinx gallery works
"""


import numpy as np
import fiducia
import matplotlib.pyplot as plt


############################################################
# Generating some data


dataX = np.arange(10)
dataY = dataX ** 2

############################################################
# Plotting the data


plt.plot(dataX, dataY)
plt.xlabel('X')
plt.ylabel('Y')
plt.title('Example data')
plt.show()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  8 09:42:24 2019

Tests for cspline.py

@author: Pawel M. Kozlowski, Myles T. Brophy
"""

# python modules
import numpy as np
import pytest
import os
import inspect

# custom modules
import fiducia
import fiducia.cspline as cspline
import fiducia.loader as loader
import scipy.sparse as sparse


class Test_splineCoords(object):
    """
    Testing class for the splineCoords and splineCorrdsInv functions
    """
    def setup_class(self):
        """Initialize parameters for tests."""
        self.energyMin = 10
        self.energyMax = 20
        self.values1 = np.arange(9) + 1
        self.energy1 = self.values1 + self.energyMin
        self.known1 = self.values1 / 10
            
    def test_splineCoords(self):
        """
        Testing for expected return value.
        """
        methodVals = cspline.splineCoords(self.energy1, self.energyMin, self.energyMax)
        testTrue = np.allclose(methodVals, self.known1, rtol=1e-16, atol=0.0)
        errStr = (f"Spline coordinate value test gives {methodVals} but "
                  f"should be {self.known1}")
        assert testTrue, errStr
        
    def test_splineCoordsInv(self):
        """Testing for expected return value."""
        methodVals = cspline.splineCoordsInv(self.known1, self.energyMin, self.energyMax)
        testTrue = np.allclose(methodVals, self.energy1,
                               rtol=1e-16, atol=0.0)
        errStr = (f"Spline coordinate inverted value test gives {methodVals} but "
                  f"should be {self.energy1}")
        assert testTrue, errStr


class Test_coeffArr(object):
    """
    Testing class for the yCoeffArr and dCoeffArr functions
    """
    def setup_class(self):
        """Initialize parameters for tests."""
        self.energyNorm = 0.5;
        self.chLen = 9
        self.trueYCoeffArr = sparse.csc_matrix([
            [0.875,0.125,0.,0.,0.,0.,0.,0.,0.,0.],
            [0.,0.5,0.5,0.,0.,0.,0.,0.,0.,0.],
            [0.,0.,0.5,0.5,0.,0.,0.,0.,0.,0.],
            [0.,0.,0.,0.5,0.5,0.,0.,0.,0.,0.],
            [0.,0.,0.,0.,0.5,0.5,0.,0.,0.,0.],
            [0.,0.,0.,0.,0.,0.5,0.5,0.,0.,0.],
            [0.,0.,0.,0.,0.,0.,0.5,0.5,0.,0.],
            [0.,0.,0.,0.,0.,0.,0.,0.5,0.5,0.],
            [0.,0.,0.,0.,0.,0.,0.,0.,0.875,0.125]])
        self.trueDCoeffArr = sparse.csc_matrix([
            [0.375,-0.125,0.,0.,0.,0.,0.,0.,0.,0.],
            [0.,0.125,-0.125,0.,0.,0.,0.,0.,0.,0.],
            [0.,0.,0.125,-0.125,0.,0.,0.,0.,0.,0.],
            [0.,0.,0.,0.125,-0.125,0.,0.,0.,0.,0.],
            [0.,0.,0.,0.,0.125,-0.125,0.,0.,0.,0.],
            [0.,0.,0.,0.,0.,0.125,-0.125,0.,0.,0.],
            [0.,0.,0.,0.,0.,0.,0.125,-0.125,0.,0.],
            [0.,0.,0.,0.,0.,0.,0.,0.125,-0.125,0.],
            [0.,0.,0.,0.,0.,0.,0.,0.,0.25,0.125]])
        
        
    def test_yCoeffArr(self):
        """Testing for expected value"""
        methodVals = cspline.yCoeffArr(self.energyNorm, self.chLen)
        testTrue = np.allclose(methodVals.toarray(), 
                               self.trueYCoeffArr.toarray(),
                               rtol=1e-16, atol=0.0)
        errStr = (f"Y coefficient array test gives {methodVals.toarray()} but "
                  f"should be {self.trueYCoeffArr.toarray()}")
        assert testTrue, errStr
        
    def test_dCoeffArr(self):
        """Testing for expected value"""
        methodVals = cspline.dCoeffArr(self.energyNorm, self.chLen)
        testTrue = np.allclose(methodVals.toarray(),
                               self.trueDCoeffArr.toarray(),
                               rtol=1e-16, atol=0.0)
        errStr = (f"D coefficient array test gives {methodVals.toarray()} but "
                  f"should be {self.trueDCoeffArr.toarray()}")
        assert testTrue, errStr


class Test_chArr(object):
    """Testing class for the __chi1Arr__ and __chi3Arr__ functions"""
    def setup_class(self):
        """Initialize parameters for test"""
        self.chLen = 6
        self.trueChi1 = sparse.csc_matrix([
            [2.,1.,0.,0.,0.,0.,0.],
            [1.,4.,1.,0.,0.,0.,0.],
            [0.,1.,4.,1.,0.,0.,0.],
            [0.,0.,1.,4.,1.,0.,0.],
            [0.,0.,0.,1.,4.,1.,0.],
            [0.,0.,0.,0.,1.,4.,1.],
            [0.,0.,0.,0.,0.,1.,2.]])
        self.trueChi3 = sparse.csc_matrix([
            [-1.,1.,0.,0.,0.,0.,0.],
            [-1.,0.,1.,0.,0.,0.,0.],
            [0.,-1.,0.,1.,0.,0.,0.],
            [0.,0.,-1.,0.,1.,0.,0.],
            [0.,0.,0.,-1.,0.,1.,0.],
            [0.,0.,0.,0.,-1.,0.,1.],
            [0.,0.,0.,0.,0.,-1.,1.]])
    
    def test___chi1Arr__(self):
        """Testing for expected return value."""
        methodVals = cspline.__chi1Arr__(self.chLen)
        testTrue = np.allclose(methodVals.toarray(), self.trueChi1.toarray(),
                               rtol=1e-16, atol=0.0)
        errStr = (f"Chi1 array test gives {methodVals.toarray()} but "
                  f"should be {self.trueChi1.toarray()}")
        assert testTrue, errStr
        
    def test___chi3Arr__(self):
        """Testing for expected return value."""
        methodVals = cspline.__chi3Arr__(self.chLen)
        testTrue = np.allclose(methodVals.toarray(), self.trueChi3.toarray(),
                               rtol=1e-16, atol=0.0)
        errStr = (f"Chi3 array test gives {methodVals.toarray()} but "
                  f"should be {self.trueChi3.toarray()}")
        assert testTrue, errStr


class Test_dToyArr(object):
    """Testing class for the dToyArr function"""
    def setup_class(self):
        self.chLen = 4
        self.trueDToyArr = [[-1.26785714,  1.60714286, -0.42857143,  0.10714286, -0.01785714],
       [-0.46428571, -0.21428571,  0.85714286, -0.21428571,  0.03571429],
       [ 0.125     , -0.75      ,  0.        ,  0.75      , -0.125     ],
       [-0.03571429,  0.21428571, -0.85714286,  0.21428571,  0.46428571],
       [ 0.01785714, -0.10714286,  0.42857143, -1.60714286,  1.26785714]]
        
    def test_dToyArr(self):
        """Testing for expected return value."""
        methodVals = cspline.dToyArr(self.chLen)
        testTrue = np.allclose(methodVals, self.trueDToyArr, 
                               rtol=1e-6, atol=0.0)
        errStr = (f"D to Y test gives {methodVals} but "
                  f"should be {self.trueDToyArr}")
        assert testTrue, errStr
        
        
# class Test_responseInterp(object):
#     """
#     Testing class for the responseInterp function.
#     """
#     def setup_class(self):
#         "Initialize parameters for testing"
#         self.energyNorm = 0.2
#         self.energyMin = 10
#         self.energyMax = 80
#         #Have to set path this way in order to work locally on pytest and
#         #on the GitLab pipeline. Not pretty.
#         self.fiduciaPath = os.path.abspath(os.path.dirname(inspect.getfile(fiducia)))
#         self.responseFilePath = os.path.join(self.fiduciaPath, "tests", "example_response.csv")
        
#         self.channels= [2, 3, 4, 5, 6, 7, 8, 9, 12, 13, 14]
#         self.responseFrame = loader.loadResponses(self.channels, self.responseFilePath)
#         self.trueResponsesInterpolated = [[8.19720e-39,0.00000e+00,0.00000e+00,0.00000e+00,2.82800e-38,
#                                            7.03465e-25,2.07858e-09,6.20039e-08,0.00000e+00,0.00000e+00,
#                                            0.00000e+00]]
        
#     def test_responseInterp(self):
#         """Testing for expected return value."""
#         methodVals = cspline.responseInterp(self.energyNorm,
#                                             self.energyMin, 
#                                             self.energyMax,
#                                             self.responseFrame,
#                                             self.channels)
#         testTrue = np.allclose(methodVals, self.trueResponsesInterpolated, 
#                                rtol=1e-13, atol=0.0)
#         errStr = (f"Response interpolation test gives {methodVals} but "
#                   f"should be {self.trueResponsesInterpolated}")
#         assert testTrue, errStr
    
    
class Test_yChiCoeff(object):
    """
    Testing class for the yChiCoeffArr and yChiCoeffArrEnergies functions
    """
    def setup_class(self):
        "Initialize parameters for testing"
        self.chLen = 4
        self.energyNorm = 0.2
        self.energyNorms = [0.2,0.3,0.5, 0.9]
        self.dToYArr = [[-1.26785714,1.60714286,-0.42857143,0.10714286,-0.01785714],
                        [-0.46428571,-0.21428571,0.85714286,-0.21428571,0.03571429],
                        [0.125,-0.75,0.,0.75,-0.125],
                        [-0.03571429,0.21428571,-0.85714286,0.21428571,0.46428571],
                        [0.01785714,-0.10714286,0.42857143,-1.60714286,1.26785714]]
    
        self.trueyChiCoeffArr = [[0.76342857,0.32342857,-0.10971429,0.02742857,-0.00457143],
                                 [-0.06342857,0.89257143,0.21371429,-0.05142857,0.00857143],
                                 [0.01714286,-0.10285714,0.92342857,0.19314286,-0.03085714],
                                 [-0.00514286,0.03085714,-0.12342857,0.97485714,0.12285714]]
        self.trueyChiCoeffArrEnergies = [[[0.76342857,0.32342857,-0.10971429,0.02742857,
                                           -0.00457143],
                                          [-0.06342857,0.89257143,0.21371429,-0.05142857,
                                           0.00857143],
                                          [0.01714286,-0.10285714,0.92342857,0.19314286,
                                           -0.03085714],
                                          [-0.00514286,0.03085714,-0.12342857,0.97485714,
                                           0.12285714]],

                                         [[0.656125,0.47925,-0.171,0.04275,
                                           -0.007125],
                                          [-0.076125,0.79975,0.342,-0.07875,
                                           0.013125],
                                          [0.020625,-0.12375,0.838,0.31275,
                                           -0.047625],
                                          [-0.006375,0.03825,-0.153,0.91675,
                                           0.204375]],

                                         [[0.45758929,0.75446429,-0.26785714,0.06696429,
                                           -0.01116071],
                                          [-0.07366071,0.56696429,0.60714286,-0.12053571,
                                           0.02008929],
                                          [0.02008929,-0.12053571,0.60714286,0.56696429,
                                           -0.07366071],
                                          [-0.00669643,0.04017857,-0.16071429,0.72767857,
                                           0.39955357]],

                                        [[0.09180357,1.02117857,-0.14271429,0.03567857,
                                          -0.00594643],
                                         [-0.01430357,0.08682143,0.97971429,-0.06267857,
                                          0.01044643],
                                         [0.00401786,-0.02410714,0.09742857,0.96139286,
                                          -0.03873214],
                                         [-0.00176786,0.01060714,-0.04242857,0.16010714,
                                          0.87348214]]]
        
    def test_yChiCoeffArr(self):
        """Testing for expected return value."""
        methodVals = cspline.yChiCoeffArr(self.energyNorm, self.chLen, self.dToYArr)
        testTrue = np.allclose(methodVals, self.trueyChiCoeffArr, 
                               rtol=1e-6, atol=0.0)
        errStr = (f"y Chi Coefficient Array test gives {methodVals} but "
                  f"should be {self.trueyChiCoeffArr}")
        assert testTrue, errStr
    
    def test_yChiCoeffArrEnergies(self):
        """Testing for expected return value."""
        methodVals = cspline.yChiCoeffArrEnergies(self.energyNorms, self.chLen, self.dToYArr)
        testTrue = np.allclose(methodVals, self.trueyChiCoeffArrEnergies, 
                               rtol=1e-5, atol=0.0)
        errStr = (f"y Chi Coefficient Energies Array test gives {methodVals} but "
                  f"should be {self.trueyChiCoeffArrEnergies}")
        assert testTrue, errStr
        
        
        
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jan 25 12:22:01 2019

FIDUCIA: Filtered Diode Unfolder (using) Cubic Spline Algorithm

DANTE spectrum deconvolver based on cubic splines method [1]. Translated
from Dan Barnak's Mathematica code.

DANTE channels are bounded by edge absorption feature (knot point) due to
filter for the respective channel. Cubic splines representing the estimated
spectrum are fitted in each spectral region bounded by knot points. The
detector signal for each channel is then equal to the response function of
the detector folded with the matrix representation of the cubic spline. A
triadiagonal matrix representation of the cubic spline equation is used to
make the problem numerically tractable. This way a matrix inversion can be
used to solve for the unknown coefficients in the cubic spline equation,
using the measured signals. These coefficients are then plugged back into
the cubic spline equation over each interval (between knot points) to
make a piecewise reconstruction of the x-ray spectrum at each time step.
    

References
----------

Cubic spline deconvolution method
[1] J. P. Knauer and N. C. Gindele. Temporal and spectral deconvolution of
data from diamond, photoconductive devices. Rev. Sci. Instrum. 75, 3714 (2004)
https://doi.org/10.1063/1.1785274

Error propagation for cubic spline deconvolution method
[2] D. L. Fehl and F. Briggs. Verification of unfold error estimates in the
unfold operator code. Rev. Sci. Instrum. 68, 890 (1997)
https://doi.org/10.1063/1.1147713
    
Useful description of cubic spline matrix representation
[3] http://mathworld.wolfram.com/CubicSpline.html

Paper comparing cubic splines unfolds to other methods
[4] D. H. Barnak, J. R. Davies, J. P. Knauer, and P. M. Kozlowski. Soft
x-ray spectrum unfold of K-edge filtered x-ray diode arrays using
cubic splines. Submitted to Review of Scientific Instruments in 2020.

@author: Pawel M. Kozlowski
"""

# python modules
import numpy as np
import matplotlib.pyplot as plt
import astropy.units as u
from astropy.constants import sigma_sb, k_B
import xarray as xr

# custom modules
from fiducia.rawProcess import (loadCorrected,
                                hysteresisCorrect,
                                align,
                                constructMeasurementFrame)
from fiducia.loader import signalsAtTime, loadResponses
from fiducia.cspline import knotSolve, reconstructSpectrum
from fiducia.visualization import plotStreak, plotTraces, plotResponse
from fiducia.response import knotFind
from fiducia.error import trapzVariance
import fiducia.pltDefaults


# listing all functions declared in this file so that sphinx-automodapi
# correctly documents them and doesn't document imported functions.
__all__ = ["simulateSignal",
           "inferRadTemp",
           "inferPower",
           "analyzeSpectrum",
           "analyzeStreak",
           "feelingLucky",
           ]


def simulateSignal():
    """
    Takes the inferred spectrum and folds it with the instrument
    function to retrieve the forward propagated signal for each
    DANTE channel.
    
    Parameters
    ----------
    
    Returns
    -------
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    return


def inferRadTemp(power, area, angle, powerUncertainty=None):
    r"""
    Gets the inferred radiation temperature by calculating in from radiated
    power through the Stefan-Boltzmann Law.
    
    Parameters
    ----------
    power: float, np.ndarray
        Total radiated power as a function of time calculated from unfolded
        spectra. See :func:`main.inferPower()`.

    area: float
        Area of emitting surface in units of mm^2. For hohlraums/halfraums,
        this is the area of the LEH.

    angle: float
        Angle between the surface area normal and the Dante line of sight in 
        degrees. Usually 37.4 degrees for hohlraums/halfraums. Must be
        between 0 and 90 degrees.

    powerUncertainty: float, np.ndarray, optional
        Uncertainty in total radiated power as a function of time calculated from unfolded
        spectra. See :func:`main.inferPower()`. The default is None.

    Returns
    -------
    tRad : numpy.ndarray
        Radiation temperature of the blackbody emitter.
        
    tRadVariance: numpy.ndarray
        Variance :math:`\sigma^2` on the radiation temperature.

    Notes
    -----
    Total x-ray flux (power) from a black body emitter is given by:
        
    .. math::
        P = \sigma_{SB} A \cos(\theta) T^4
        
        
    Where P = power, :math:`\sigma_{SB}` = Stefan-Boltzmann constant, A is the
    area of radiating surface, :math:`\theta` is the viewing angle between the
    surface area normal and the Dante line-of-sight, T is the radiation
    temperature of the black body emitter.
    
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    # limit viewing angles to 90 degrees.
    if not 0 <= angle <= 90:
        raise Exception("Angle must be between 0 and 90 degrees.")
    
    # when power uncertainty is not given, it is set to zero.
    if powerUncertainty is None:
        if type(power) is np.ndarray:
            powerUncertainty = np.zeros(power.shape)
        else:
            powerUncertainty = 0
    
    cosAngle = np.cos(np.deg2rad(angle))
    #Stefan-Boltzmann constant
    sb = sigma_sb.to(u.GW/(u.mm ** 2 * u.K ** 4))
    sigma_sb_unitless = sb.value
    kb = k_B.to(u.eV/u.K).value
    denom = area * cosAngle * sigma_sb_unitless
    # value to be quadratic rooted for getting radiation temperature
    val2Root = power / denom
    # quadratic rooting to get Trad in Kelvins, and converting from
    # Kelvins to electronvolts.
    tRad = np.power(val2Root, 0.25) * kb
    tRadVariance = 0.0
    if np.shape(powerUncertainty) == np.shape(power):
#        tRadVariance = kb ** 2 / np.sqrt(denom) * (0.25 * power ** (-0.75) * powerUncertainty) ** 2
        tRadVariance = (0.25 * tRad * powerUncertainty / power) ** 2
        # cleaning up cases where we get nan because input power is zero
        tRadVariance = np.nan_to_num(tRadVariance)
    return tRad, tRadVariance


def inferPower(energies, spectra, spectraUncertainty=None):
    r"""
    Gets the inferred total radiation power as a function of time.
    
    Parameters
    ----------
    energies: numpy.ndarray
        Photon energies corresponding to input spectrum
        
    spectra: numpy.ndarray
        Spectral Flux values as a function of photon energy in units of
        (GW/sr/eV)
    
    Returns
    -------
    power: numpy.ndarray
        Total x-ray power (flux) as a function of time.
        
    powerVariance: numpy.ndarray
        Variance :math:`\sigma^2` on total x-ray power.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    if spectraUncertainty is None:
        spectraUncertainty = np.zeros(spectra.shape)

    timesLength = np.shape(spectra)[1]
    power = np.zeros(timesLength)
    powerVariance = np.zeros(timesLength)
    for idx in range (0, timesLength):
        power[idx] = np.trapz(y = spectra[:, idx], x=energies[:, idx])
        powerVariance[idx] = trapzVariance(spectraUncertainty[:, idx],
                                           x=energies[:, idx])
    return power, powerVariance


def analyzeSpectrum(channels,
                    knots,
                    detArr,
                    detArrBoundaryCol,
                    detArrVarianceBoundaryCol,
                    detArrInv,
                    stdDetArrInv,
                    measurementFrame,
                    time,
                    signalsUncertainty = None,
                    yGuess=0,
                    boundary="y0",
                    nPtsIntegral=100,
                    nPtsSpectrum=100,
                    plotSignal=False,
                    plotKnots=False,
                    plotSpectrum=True):
    r"""
    Given the response function file and the DANTE measurement data file,
    run cubic spline analysis to reconstruct spectrum for a given time.
    
    
    Parameters
    ----------
    channels: list, numpy.ndarray
        List or array of relevant DANTE channel numbers.
    
    responseFrame: pandas.core.frame.DataFrame
        Pandas dataFrame containing response functions for each DANTE
        channel. See loadResponses().
        
    knots: list, numpy.ndarray
        List or array of knot point photon energy value. See knotFind().
    
    detArr : xarray.DataArray
        Matrix representing the spectrally integrated folding of the detector
        response with a cubic spline interpolation of the x-ray spectrum.
        2D array of channels and knot points of shape (n, n).
 
    detArrBoundaryCol : xarray.DataArray
        Column of cublic spline matrix corresponding to the knots at the boundary
        chosen with `boundary`.
 
    detArrVarianceBoundaryCol: xarray.DataArray
        Column of variances in the cublic spline matrix corresponding to the 
        knots at the boundary chosen with `boundary`.   
 
    detArrInv : xarray.DataArray
        Inversion of detArr, with the column corresponding to boundary removed so detArr is invertible.
   
    stdDetArrInv : xarray.DataArray
        Array of the standard deviation of each element in detArrInv based on variance
        using the `responseUncertaintyFrame` propagated with Monte Carlo. 
    
    measurementFrame: pandas.core.frame.DataFrame
        Pandas dataframe containing DANTE measurement data. See
        readDanteData() and readDanProcessed().
        
    time: float
        Time for which we want DANTE signals (in ns).

    signalsUncertainty: numpy.ndarray, optional
        One dimensional array with each element corresponding to the uncertainty
        each signal. The default is None.        
        
    yGuess: float, optional
        Guess for position of boundary knot point. Default 0.
        
    boundary: str, optional
        Choose whether yGuess corresponds to :math:`y_0` (lowest photon energy) or
        :math:`y_{n+1}` (highest photon energy) boundary condition. This should
        correspond to the photon energy value given in knots. Options are `y0`
        or `yn+1`. Default 'y0'.
    
    nPtsIntegral: int, optional
        Number of points used in computing the integral. Default is 100.
        
    nPtsSpectrum: int, optional
        Number of points to use in reconstructing the spectrum. Default is 100.
    
    plotKnots: Bool, optional
        Flag for plotting the Dante signal at the given time across all
        channels. Default is False.
    
    plotKnots: Bool, optional
        Flag for plotting just the solved knot points. Default is False.
        
    plotSpectrum: Bool, optional
        Flag for plotting the unfolded spectrum. Default is True.
    
    Returns
    -------    
    knotsYAll: numpy.ndarray
        Array of knot point intensity values with yGuess appended.
        See knotSolve() and analyzeSpectrum().

    knotsYVariance: numpy.ndarray
        Array of knot point intensity uncertainty values with yGuess appended.
        See knotSolve() and analyzeSpectrum().
    
    photonEnergies: numpy.ndarray
        Photon energy axis of unfolded spectrum.
        
    intensities: numpy.ndarray
        Spectral intensity axis of unfolded spectrum.
        
    intensitiesVariance: numpy.ndarray
        Uncertaitny (1 :math:`\sigma`) on spectral intensity values.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    # Select a particular time step for generating x-ray spectrum from DANTE
    signals = signalsAtTime(time=time,
                            measurementFrame=measurementFrame,
                            channels=channels,
                            plot=plotSignal,
                            method="interp")
    # solve knots and plot
    # if signalsUncertainty is None:
    #     signalsUncertainty = np.ones(signals.shape)
        
    knotsY, knotsYVariance = knotSolve(signals,
                                       detArr, detArrBoundaryCol,
                                       detArrVarianceBoundaryCol,
                                       detArrInv, stdDetArrInv,
                                       signalsUncertainty,
                                       yGuess=yGuess,
                                       npts=nPtsIntegral)
    #Root the variance to get the uncertainty (sigma)
    knotsYUncertainty = np.sqrt(knotsYVariance)
    
    # prepending y0 guess to form full array of y knot values equal in
    # length to knot energy values
    if boundary == "y0":
        knotsYAll = np.append([yGuess], knotsY)
        knotsYUncertaintyAll = np.append([yGuess], knotsYUncertainty)

    elif boundary == "yn+1":
        knotsYAll = np.append(knotsY, [yGuess])
        knotsYUncertaintyAll = np.append(knotsYUncertainty, [yGuess])
    else:
        raise Exception(f"No method found for boundary {boundary}.")
    
    # making simple plot at this intermediate step to get rough idea
    # of the shape of the spectrum
    if plotKnots:
            plt.plot(knots, knotsYAll)
            plt.xlabel("Energy (eV)")
            plt.ylabel("X-ray spectrum")
            plt.title("cubic spline knots")
            plt.show()
    
    # reconstruct spectrum using inferred knot points
    chLen = len(channels)
    photonEnergies, intensities, intensitiesVariance = reconstructSpectrum(chLen,
                                                                           knots,
                                                                           knotsYAll,
                                                                           knotsYUncertaintyAll,
                                                                           nPtsSpectrum,
                                                                           plot=plotSpectrum)
    photonEnergiesPlus, intensitiesPlus, _ = reconstructSpectrum(chLen,
                                                                 knots,
                                                                 knotsYAll+knotsYUncertaintyAll,
                                                                 knotsYUncertaintyAll,
                                                                 nPtsSpectrum,
                                                                 plot=False)
    photonEnergiesMinus, intensitiesMinus, _ = reconstructSpectrum(chLen,
                                                                   knots,
                                                                   knotsYAll-knotsYUncertaintyAll,
                                                                   knotsYUncertaintyAll,
                                                                   nPtsSpectrum,
                                                                   plot=False)
    #subtract to just get the delta from intensities. plot_line_shaded just wants the difference
    intensitiesPosDiff = intensitiesPlus-intensities
    intensitiesNegDiff = intensities - intensitiesMinus
    if plotSpectrum:
        fiducia.pltDefaults.plot_line_shaded(photonEnergies, 
                                             intensities, 
                                             intensitiesPosDiff, 
                                             intensitiesNegDiff,)
        plt.xlabel("Photon Energy (eV)")
        plt.ylabel("Spectrum (GW/sr/eV)")
        plt.title("Spectra with error bars")
        plt.show()    
    return knotsYAll, knotsYVariance, photonEnergies, intensities, intensitiesVariance


def analyzeStreak(channels,
                  responseFrame,
                  knots,
                  detArr,
                  detArrBoundaryCol,
                  detArrVarianceBoundaryCol,
                  detArrInv, 
                  stdDetArrInv,
                  measurementFrame,
                  timeStart,
                  timeStop,
                  timeStep,
                  signalsUncertainty=None,
                  yGuess=0,
                  boundary="y0",
                  nPtsIntegral=100,
                  nPtsSpectrum=100):
    r"""
    Given the response function file and the DANTE measurement data file,
    run cubic spline analysis to reconstruct spectrum for a given time.
    
    Parameters
    ----------
    channels: list, numpy.ndarray
        List or array of relevant DANTE channel numbers.
    
    responseFrame: pandas.core.frame.DataFrame
        Pandas dataFrame containing response functions for each DANTE
        channel. See loadResponses().
        
    knots: list, numpy.ndarray
        List or array of knot point photon energy value. See knotFind().
    
    detArr : xarray.DataArray
        Matrix representing the spectrally integrated folding of the detector
        response with a cubic spline interpolation of the x-ray spectrum.
        2D array of channels and knot points of shape (n, n).
 
    detArrBoundaryCol : xarray.DataArray
        Column of cublic spline matrix corresponding to the knots at the boundary
        chosen with `boundary`.
 
    detArrVarianceBoundaryCol: xarray.DataArray
        Column of variances in the cublic spline matrix corresponding to the 
        knots at the boundary chosen with `boundary`.   
        
    detArrInv : xarray.DataArray
        Inversion of detArr, with the column corresponding to boundary removed so detArr is invertible.
   
    stdDetArrInv : xarray.DataArray
        Array of the standard deviation of each element in detArrInv based on variance
        using the `responseUncertaintyFrame` propagated with Monte Carlo. 
    
    measurementFrame: pandas.core.frame.DataFrame
        Pandas dataframe containing DANTE measurement data. See
        :func:`loader.readDanteData` and :func:`readDanProcessed`.
        
    timeStart: float
        Start time for producing temporally streaked DANTE spectra (in ns).
        
    timeStop: float
        End time for producing temporally streaked DANTE spectra (in ns).
        
    timeStep: float
        Time step size for producing temporally streaked DANTE spectra (in ns).
        
    signalsUncertainty: numpy.ndarray, optional
        One dimensional array with each element corresponding to the uncertainty
        each signal. The default is None. 
    
    yGuess: float, optional
        Guess for position of boundary knot point. Default is 1e-77.
        
    boundary: str, optional
        Choose whether yGuess corresponds to :math:`y_0` (lowest photon energy) or
        :math:`y_{n+1}` (highest photon energy) boundary condition. This should
        correspond to the photon energy value given in knots. Options are `y0`
        or `yn+1`. Default 'y0'.
    
    nPtsIntegral: int, optional
        Number of points used in computing the integral. Default is 100.
        
    nPtsSpectrum: int, optional
        Number of points to use in reconstructing the spectrum. Default is 100.
    
    Returns
    -------
    times
    
    
    energies
    
    
    spectra
    
    
    spectraVariance
        
    
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
        
    """
    chLen = len(channels)
    # generating evenly spaced timesteps
    times = np.arange(timeStart, timeStop + timeStep, timeStep)
    timesLen = len(times)
    # initializing array for holding spectral values
    spectralPts = chLen * nPtsSpectrum - chLen
    energies = np.zeros((spectralPts, timesLen))
    spectra = np.zeros((spectralPts, timesLen))
    #TODO implement spectraVariance once DotVariance works on N-D arrays
    spectraVariance = np.zeros((spectralPts, timesLen))
    
    for idt, time in enumerate(times):
        # process the spectrum for the given time
        results = analyzeSpectrum(channels,
                                  knots,
                                  detArr, 
                                  detArrBoundaryCol,
                                  detArrVarianceBoundaryCol,
                                  detArrInv, 
                                  stdDetArrInv,
                                  measurementFrame,
                                  time,
                                  signalsUncertainty=signalsUncertainty,
                                  yGuess=yGuess,
                                  boundary=boundary,
                                  nPtsIntegral=100,
                                  nPtsSpectrum=100,
                                  plotKnots=False,
                                  plotSpectrum=False)
        knotsYAll, knotsYVariance, photonEnergies, intensities, intensitiesVariance = results
        # including unfolding spectrum for time step into array of spectra
        # for all time steps
        energies[:, idt] = photonEnergies
        spectra[:, idt] = intensities
        spectraVariance[:, idt] = intensitiesVariance
        print(f"Completed time step {time} ns.")    
    # plotting streaked spectrum
    plotStreak(times, energies, spectra)
    return times, energies, spectra, spectraVariance


def feelingLucky(dataFile,
                 attenuatorsFile,
                 offsetsFile,
                 responseFile,
                 csplineDatasetFile,
                 channels,
                 area,
                 angle,
                 signalsUncertainty=None,
                 peaksNum=2):
    r"""
    Attempt processing dante signals given dante data file and calibration
    files using sensible defaults.
    
    
    Parameters
    ----------
    dataFile: str
        Full path to the Dante .dat file containing dante signals from LLE
        site.
        
    attenuatorsFile: str
        Full path to file containing attenuator serial numbers and
        corresponding attenuation factors.
        
    offsetsFile: str
        Full path to file containing oscilloscope channel offsets in
        time and voltage.
        
    responseFile: str
        Full path and filename of .csv file containing DANTE respones
        functions corresponding to dataFile.
    
    csplineDatasetFile : str
        File pointing to the path of the saved dataset containing ''detArr'',
        ''detArrBoundaryCol'',  ''detArrInv'', and ''stdDetArrInv''. 
        See :func:'error.analyzeSpectrumUncertainty()'.
  
    channels: list, numpy.ndarray
        List or array of relevant channels for which to apply analysis.
        
    area: float
        Area of emitting surface in units of mm^2. For hohlraums/halfraums,
        this is the area of the LEH. Used in Trad calculation.
        
    angle: float
        Angle between the surface area normal and the Dante line of sight in 
        degrees. Usually 37.4 degrees for hohlraums/halfraums. Used in Trad 
        calculation.
    
    signalsUncertainty: numpy.ndarray, optional
        One dimensional array with each element corresponding to the uncertainty
        each signal. The default is None.
    
    Returns
    -------
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    #load cspline matrix related files
    csplineDataset =  xr.open_dataset(csplineDatasetFile).load()
    detArr = csplineDataset['detArr']
    detArrBoundaryCol = csplineDataset['detArrBoundaryCol']
    detArrVarianceBoundaryCol = csplineDataset['detArrVarianceBoundaryCol']
    detArrInv = csplineDataset['detArrInv']
    stdDetArrInv = csplineDataset['stdDetArrInv']
    #close out of the files
    csplineDataset.close()
    
    #find what boundary was used when calculating cspline matrices
    boundary = csplineDataset.attrs['boundary']
    
    # loading data and applying corrections.
    timesFrame, dfAtten, onChList, hf, dfVolt = loadCorrected(danteFile=dataFile,
                                                              attenuatorsFile=attenuatorsFile,
                                                              offsetsFile=offsetsFile,
                                                              plot=True)
    
    # removing hysteresis and background with a polynomial fit
    dfPoly = hysteresisCorrect(timesFrame=timesFrame,
                               df=dfAtten,
                               channels=channels,
                               order=5,
                               prominence=0.2,
                               width=10,
                               avgMult=1)
    
    # aligning signals to peak
    # aligning to 1e-9 seconds by default.
    # looking for 2 peaks and aligning to the first (0th index) peak.
    timesAligned = align(timesFrame=timesFrame,
                         df=dfPoly,
                         channels=channels,
                         peaksNum=peaksNum,
                         peakAlignIdx=0,
                         referenceTime=1e-9,
                         prominence=0.01,
                         width=10,
                         avgMult=1.5)
    
    # constructing dataframe for passing to analyzeStreak()
    measurementFrame = constructMeasurementFrame(timesFrame=timesAligned,
                                                 df=dfPoly,
                                                 channels=channels)
        
    # testing plot traces
    plotTraces(channels, measurementFrame, scale='log')
    
    # loading dante responses
    responseFrame = loadResponses(channels, responseFile)

    # getting knots from response function edges
    # currently not forcing any knots
    knots = knotFind(channels=channels,
                      responseFrame=responseFrame,
                      forceKnot=np.array([]),
                      knotBoundary=0,
                      boundary=boundary)
    
    # print(f"Channels: {channels}")
    # print(f"Channels length: {len(channels)}")
    # print(f"Knot energies (eV): {knots}")
    # print(f"Knot energies length: {len(knots)}")
    
    # plotting all relevant response functions
    # plotResponse(channels=channels, 
    #              responseFrame=responseFrame,
    #              knots=knots)
    
    # unfolding the spectrum for a particular time step
    # currently unfolding at 1 ns
    time = 1.0
    spectrumResults = analyzeSpectrum(channels, 
                                      knots, 
                                      detArr,
                                      detArrBoundaryCol,
                                      detArrVarianceBoundaryCol,
                                      detArrInv,
                                      stdDetArrInv,
                                      measurementFrame,
                                      signalsUncertainty=signalsUncertainty,
                                      time=time,
                                      yGuess=0,
                                      boundary=boundary,
                                      plotKnots=True)
    knotsYAll, knotsYVariance, photonEnergies, intensities, intensitiesVariance = spectrumResults
    
    # unfolding time resolved spectra
    timeStart = -1
    timeStop = 4
    timeStep = 0.1
    streakResults = analyzeStreak(channels,
                                  responseFrame,
                                  knots,
                                  detArr, 
                                  detArrBoundaryCol,
                                  detArrVarianceBoundaryCol,
                                  detArrInv, 
                                  stdDetArrInv,
                                  measurementFrame,
                                  timeStart,
                                  timeStop,
                                  timeStep,
                                  signalsUncertainty=signalsUncertainty,
                                  yGuess=0,
                                  boundary=boundary,
                                  nPtsIntegral=100,
                                  nPtsSpectrum=100)
    times, energies, spectra, spectraVariance = streakResults


    spectraUncertainty = np.sqrt(spectraVariance)
    power, powerVariance = inferPower(energies, spectra, spectraUncertainty)
    plt.scatter(times, power)
    plt.xlabel("Time (ns)")
    plt.ylabel("Radiated Power (GW/sr)")
    plt.show()
    powerUncertainty = np.sqrt(powerVariance)
    fiducia.pltDefaults.plot_line_shaded(times, power, powerUncertainty)
    plt.xlabel("Time (ns)")
    plt.ylabel("Radiated Power (GW/sr)")
    axes = plt.gca()
    axes.set_ylim([0,None])
    plt.show()
    
    tRad, tRadVariance = inferRadTemp(power, area, angle, powerUncertainty)
    plt.scatter(times, tRad)
    plt.xlabel("Time (ns)")
    plt.ylabel("Radiation Temperature (eV)")
    plt.show()
    tRadUncertainty = np.sqrt(tRadVariance)
    fiducia.pltDefaults.plot_line_shaded(times, tRad, tRadUncertainty)
    plt.xlabel("Time (ns)")
    plt.ylabel("Radiation Temperature (eV)")
    axes = plt.gca()
    axes.set_ylim([0,None])
    plt.show()
#    return None
    return times, energies, spectra, power, tRad



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  8 09:20:37 2019

Utilities for loading DANTE measurement and response function data.

@author: Pawel M. Kozlowski
"""

# python modules
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import scipy.sparse as sparse


# custom modules
from fiducia.misc import find_nearest
import fiducia.pltDefaults


# listing all functions declared in this file so that sphinx-automodapi
# correctly documents them and doesn't document imported functions.
__all__ = ["cleanupHeader",
           "loadResponses",
           "loadResponseUncertainty",
           "readDanProcessed",
           "signalsAtTime",
           "signalInt",
           "readDanteData",
           ]


def __responseName__(channelNum):
    r"""
    Convenience function for generating response function file name
    given the DANTE channel number.
    
    Parameters
    ----------
    channelNum: int
        DANTE channel number
    
    Returns
    -------
    fileName: str
        The file name of the response file for channel 'channelNum'
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    baseName = 'DanteDataS_1_'
    extension = '.dat'
    fileName = baseName + str(channelNum) + extension
    return fileName


def __readResponse__(channelNum, directory):
    r"""
    Read a single DANTE channel response function file given the channel
    number and path to the directory containing the response function files.
    
    Parameters
    ----------
    channelNum: int
        DANTE channel number
        
    directory: str
        Path to channel response function files
        
    Raises
    ------
    Exception
        If file does not exist.
    
    Returns
    -------
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    try:
        fileName = __responseName__(channelNum)
        responseArr = np.loadtxt(directory + fileName)
        return responseArr
    except IOError:
        # file doesnt exist, return dummy data
        Exception(f"File {fileName} does not exist!")


def cleanupHeader(dataFrame):
    r"""
    Strip whitespace and rename DataFrame headers.

    Parameters
    ----------
    dataFrame : pandas.core.frame.DataFrame
        DataFrame to be cleaned.

    Returns
    -------
    cleanedDataFrame : pandas.core.frame.DataFrame
        DataFrame with stripped and renamed channel headers.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------

    """
    # stripping whitespace from header names
    colNames = dataFrame.columns
    colNamesStripped = [header.strip() for header in colNames]
    renameStripDict = {colName: colNamesStripped[idx] for idx, colName in enumerate(colNames)}
    df1 = dataFrame.rename(columns=renameStripDict)
    # renaming channel headers from e.g. Ch2 to just 2
    allChannels = np.arange(18) + 1
    renameDict = {}
    for channel in allChannels:
        renameDict['Ch' + str(channel)] =  channel

    cleanedDataFrame = df1.rename(columns=renameDict)
    return cleanedDataFrame


def loadResponses(channels, fileName, solid=True):
    r"""
    Load DANTE measurement data from files given the channels and path to the
    directory containing the response function files. Returns a dataframe
    with the data.
    
    Parameters
    ----------
    channels: list, numpy.ndarray
        List or array of relevant channels
        
    fileName: str
        Full path and filename of .csv file containing DANTE respones
        functions.
    
    solid: Bool, optional
        Includes solid angle in response function value if true. The default is true.
        
    Returns
    -------
    responseFrame:  pandas.core.frame.DataFrame
        DataFrame with the response function data for the 'channels' requested
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """

    solidAngles = fiducia.misc.solidAngles
    # loading all the response functions
    dataFrame = pd.read_csv(fileName)
    #clean headers
    cleanedFrame = cleanupHeader(dataFrame)
    # filtering for channels we care about
    # for this particular shot
    colFilter = ['Energy(eV)'] + channels
    responseFrame = cleanedFrame[colFilter].copy()
    # convert energy column from strings to floats (if necessary)
    if type(responseFrame['Energy(eV)'][0]) == str:
        energyFloats = responseFrame['Energy(eV)'].str.replace(',', '').astype(float)
        responseFrame.loc[:,'Energy(eV)'] = energyFloats
    else:
        energyFloats = responseFrame['Energy(eV)'].astype(float)
        responseFrame.loc[:,'Energy(eV)'] = energyFloats

    if solid:
        for chan in channels:
            #multiply each element by the corresponding channel solid angle
            responseFrame.loc[:, chan] *= solidAngles[chan-1]
            #save metadata that we already include solid angle
            responseFrame.solid = True

    return responseFrame


def loadResponseUncertainty(responseFrame, fileName):
    r"""
    Load uncertainty percentages into a DataFrame.
    
    Parameters
    ----------
    responseFrame: pandas.core.frame.DataFrame
        DataFrame to base the respones uncertainty frame on. 
        
    fileName: str
        Full path and filename of .csv file containing DANTE response uncertainty
        percentages functions.

    Returns
    -------
    responseUncertaintyFrame : pandas.core.frame.DataFrame
        DataFrame with each column being a channel and each element being the
        channel's uncertainty percentage. Extended to match the photon energy
        range in the response frame.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    #read raw data
    channelUncertaintyFrame = pd.read_csv(fileName)
    #clean headers
    cleanedFrame = cleanupHeader(channelUncertaintyFrame)
    # filtering for columns we care about
    channels = list(responseFrame.columns.values)
    channels.remove('Energy(eV)')
    responseUncertaintyFrame = responseFrame.copy()
    #go through each column and fill each element with corresponding uncertainty
    for chan in channels:
        #multiply each responseFrame element by the percent uncertainty
        responseUncertaintyFrame.loc[:, chan] *= cleanedFrame.loc[0, chan]/100
    return responseUncertaintyFrame


def readDanProcessed(channels, directory):
    r"""
    Loads DANTE measurement data from files given the channels and path to the
    directory containing the reduced and aligned DANTE data. Returns a dataframe
    with the data.
    Note that this is *not* for raw data. It is for reading DANTE signals
    that have already been processed by Dan Barnak's scripts.
    
    Parameters
    ----------
    channels: list, numpy.ndarray
        List or array of relevant channels
        
    directory: str
        Path to channel response function files
    
    Returns
    -------
    dataFrame : pandas.core.frame.DataFrame
        Dataframe of aligned signals from Dan's analysis.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    
    """
    # initialize finder for length of data array for longest channel
    longestChLen = 0
    # load all the arrays into a dict
    responseDict = {}
    for channel in channels:
        responseData = __readResponse__(channel, directory)
        responseDict[channel] = responseData
        # getting length of array for longest channel
        responseLen = len(responseData)
        if responseLen > longestChLen:
            longestChLen = responseLen
    # initialize pandas dataframe
    indices = np.arange(longestChLen)
    dataFrame = pd.DataFrame(index=indices)
    # add index axis label
    dataFrame.index.name = 'indices'
    # write channel response data into dataframe
    for channel in channels:
        chData = responseDict[channel]
        # photon energy
        dataFrame['Time' + str(channel)] = chData[:,0]
        # channel response
        dataFrame['Signal' + str(channel)] = chData[:,1]
    return dataFrame


def signalsAtTime(time,
                  measurementFrame,
                  channels,
                  plot=False,
                  method="interp"):
    r"""
    Get DANTE signals from each channel at a particular time. Default is
    to return an interpolated value of the signal at the given time.
    Alternatively, this function can return the nearest value in
    the signal data array for the given time.
    
    
    Parameters
    ----------
    time: float
        Time for which we want DANTE signals (in ns).
        
    measurementFrame: pandas.core.frame.DataFrame
        Pandas dataframe containing DANTE measurement data. See
        readDanteData() and readDanProcessed().
        
    plot: Bool
        When True, plots DANTE signals vs channel index at a particular time.
        
    method: str
        Either 'nearest' or 'interp'. 'nearest' finds the nearest point in the
        DANTE signal to the given time. 'interp' returns an interpolated
        signal value for the given time. Default is 'interp'.
    
    Returns
    -------
    signals : numpy.ndarray
        Dante signals for each channel at a particular time step.
    
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    chLen = len(channels)
    signals = np.zeros(chLen)
    for idx, channel in enumerate(channels):
        if method == "nearest":
            timeIdx, _ = find_nearest(array=measurementFrame['Time' + str(channel)],
                                      value=time)
            signals[idx] = measurementFrame['Signal' + str(channel)][timeIdx]
        elif method == "interp":
            signals[idx] = np.interp(x=time,
                                     xp=measurementFrame['Time' + str(channel)],
                                     fp=measurementFrame['Signal' + str(channel)])
        else:
            raise Exception(f"Method {method} not found!")
        
    if plot:
        plt.scatter(channels, signals)
        plt.xticks(channels)
        plt.xlabel('DANTE channel')
        plt.ylabel('Signal (V)')
        plt.title(f'DANTE signals @ t = {time} ns')
        plt.show()
    return signals


def signalInt(channels, measurementFrame, tStart, tEnd):
    r"""
    Get time-integrated Dante signals for a specified time interval. Used in
    getting time-integrated spectrum from the unfold.
    
    Parameters
    ----------
    measurementFrame: pandas.core.frame.DataFrame
        Pandas dataframe containing DANTE measurement data. See
        loadDanteData().
    tStart: float
        Lower bound for time integration.
        
    tEnd: float
        Upper bound for time integration
    
    Returns
    -------
    signalInt: numpy.ndarray
        Time integrated Dante signals for each channel.
    
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    from scipy import integrate
    
    chLen = len(channels)
    signalInt = np.zeros(chLen)
    for idx, channel in enumerate(channels):
        timeseries = measurementFrame['Time' + str(channel)]
        chanseries = measurementFrame['Signal' + str(channel)]
        timeIdx1, _ = find_nearest(array=timeseries, value=tStart)
        timeIdx2, _ = find_nearest(array=timeseries, value=tEnd)
        signalInt[idx] = integrate.simps(y=chanseries[timeIdx1:timeIdx2],
                                         x=timeseries[timeIdx1:timeIdx2])
    return signalInt


def readDanteData(filePath):
    r"""
    Reads Dante .dat file and returns header info and channel signals
    as two separate pandas dataframes.
    
    Parameters
    ----------
    filePath: str
        Full path to the Dante .dat file.
    
    Returns
    -------
    headerFrame: pandas.core.frame.DataFrame
        Header of Dante data file. This typically include information
        about the various components used in each Dante channel, such
        as oscilloscopes, XRDs, etc.
    
    dataFrame: pandas.core.frame.DataFrame
        Dante data.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    # reading the entire dante file
    dataAndHeaderFrame = pd.read_csv(filePath, sep='\t', header=None)
    # generating dante dataframe header names
    headerNames1st = [str(num) for num in np.arange(18) + 1]
    headerNames2nd = [str(num) + ' bkg' for num in np.arange(18) + 1]
    headerName = headerNames1st + headerNames2nd
    # replacing dataframe column header names for more intuitive access
    dataAndHeaderFrame.columns = headerName
    # splitting into header frame and measurement data frame
    headerLen = 18
    headerFrame = dataAndHeaderFrame[:][:headerLen]
    dataFrame = dataAndHeaderFrame[:][headerLen:]
    # replacing row names for header frame
    indexNamesReplace = {0:'Signal Cable',
                         1:'Attenuator 1',
                         2:'Attenuator 2',
                         3:'Attenuator 3',
                         4:'Attenuator 4',
                         5:'Jumper Cable',
                         6:'XRD SN',
                         7:'Mirror SN',
                         8:'Filter 1 SN',
                         9:'Filter 2 SN',
                         10:'Filter 3 SN',
                         11:'Fiducial T',
                         12:'Scope type',
                         13:'Full scale Hor time',
                         14:'#Hor Pts',
                         15:'Full Scale Vert mV',
                         16:'HV bias for XRDs',
                         17:'(unused field)'}
    headerFrame.rename(index=indexNamesReplace, inplace=True)
    return headerFrame, dataFrame



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  8 10:49:45 2019

Tests for visualization.py

@author: Pawel M. Kozlowski
"""

# python modules
import numpy as np
import pytest

# custom modules
import fiducia.visualization as vis


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  8 10:49:17 2019

Utilities for visualizing DANTE data.

@author: Pawel M. Kozlowski
"""

# python modules
import numpy as np
import matplotlib.pyplot as plt

# custom modules
import fiducia.pltDefaults


# listing all functions declared in this file so that sphinx-automodapi
# correctly documents them and doesn't document imported functions.
__all__ = ["plotResponse",
           "plotTraces",
           "plotStreak",
           "signalImg"]


def plotResponse(channels,
                 responseFrame,
                 knots,
                 solid=True,
                 title='Dante Response Functions'):
    r"""
    Plots response function curves with knot locations identified as
    vertical dashed lines.
    
    channels: list, numpy.ndarray
        List or array of relevant channels
    
    responseFrame: pandas.core.frame.DataFrame
        Pandas dataFrame containing response functions for each DANTE
        channel. See loadResponses().
        
    knots: list, numpy.ndarray
        List or array of knot point photon energy value. See knotFind().
        
    solid: Bool
        Includes solid angle in response function value if true.
        Necessary for plotting responses with correct units.
        
    Parameters
    ----------
    
    Returns
    -------
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
        
    """
    # getting sensible figure bounds
    yMax = np.max(np.max(responseFrame[channels]))
    yMin = yMax / 1e11
    xMin = 1e1
    xMax = 1e5
    # plotting figure
    fig = plt.figure()
    for idx, channel in enumerate(channels):
        plt.loglog(responseFrame['Energy(eV)'],
                   responseFrame[channel],
                   label=str(channel))
    for idx, _ in enumerate(knots):
        plt.loglog((knots[idx], knots[idx]), (yMin, yMax), '--', color='grey')
    plt.xlim((xMin, xMax))
    plt.ylim((yMin, yMax))
    plt.legend(frameon=False,
               labelspacing=0.001,
               borderaxespad=0.1)
    plt.xlabel('Energy (eV)')
    if solid:
        plt.ylabel('Response (V/GW/sr)')
    else:
        plt.ylabel('Response (V/GW)')        
    plt.title(title)
    plt.show()
    return


def plotTraces(channels, measurementFrame, scale='regular'):
    r"""
    Given a dataframe of Dante channel data, plot all the signal
    traces onto a single plot.
    
    channels: list, numpy.ndarray
        List or array of relevant channels
        
    measurementFrame: pandas.core.frame.DataFrame
        Pandas dataframe containing DANTE measurement data. See
        readDanteData() and readDanProcessed().
        
    Parameters
    ----------
    
    Returns
    -------
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    if scale == 'regular':
        for channel in channels:
            plt.plot(measurementFrame['Time' + str(channel)],
                     measurementFrame['Signal' + str(channel)],
                     label=str(channel))
        plt.legend(frameon=False,
                   labelspacing=0.001,
                   borderaxespad=0.1)
        plt.xlabel('Time (ns)')
        plt.ylabel('Signal (V)')
        plt.title('Dante Measurement Data')
        plt.show()
    elif scale == 'log':
        for channel in channels:
            plt.semilogy(measurementFrame['Time' + str(channel)],
                         measurementFrame['Signal' + str(channel)],
                         label=str(channel))
        plt.legend(frameon=False,
                   labelspacing=0.001,
                   borderaxespad=0.1)
        plt.xlabel('Time (ns)')
        plt.ylabel('Signal (V)')
        plt.title('Dante Measurement Data')
        plt.show()
    else:
        raise Exception(f"No such method for scale: {scale}")
    return


def uniformStreak(times, energies, spectra):
    r"""
    Convert streak spectra from an irregularly spaced energy sampling
    to uniformly spaced energies via linear interpolation. This is
    primarily useful for visualizing the streaked spectrum in plotStreak().
    
    times : numpy.ndarray
        1D array of times corresponding to streaked spectrum
        
    energies : numpy.ndarray
        2D array of energies which are nonuniformly spaced. Each column
        corresponds to the photon energy axis for a particularly time.
        Typically all the columns are identical.
        
    spectra : numpy.ndarray
        2D array of streaked spectra corresponding to energies array.
    """
    # generate a uniform energy axis
    energyMin = int(np.min(energies)) # eV
    energyMax = int(np.max(energies)) # eV
    energyStep = 1 # eV
    energyNew = np.arange(start=energyMin, stop=energyMax, step=energyStep)
    newShape = (len(energyNew), len(times))
    energiesNew = np.zeros(newShape)
    spectraNew = np.zeros(newShape)
    for idx, time in enumerate(times):
        energyOld = energies[:, idx]
        spectrumOld = spectra[:, idx]
        spectrumNew = np.interp(x=energyNew, xp=energyOld, fp=spectrumOld)
        # storing values
        energiesNew[:, idx] = energyNew
        spectraNew[:, idx] = spectrumNew
    return energiesNew, spectraNew


def plotStreak(times, energies, spectra):
    r"""
    Plot streak of unfolded Dante spectra. See analyzeStreak().
    
    Parameters
    ----------
    times: numpy.ndarray
        Array of times for which the unfold was analyzed.
        
    energies: numpy.ndarray
        Array of photon energies corresponding to the unfolded spectra.
        
    spectra: numpy.ndarray
        The unfolded spectral intensities as a 2D array. See analyzeStreak().
    
    Returns
    -------
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    energiesInterp, spectraInterp = uniformStreak(times, energies, spectra)
    extent=[energiesInterp[0,0],
            energiesInterp[-1,0],
            times[0],
            times[-1]]
    
    plt.imshow(spectraInterp.transpose(),
               extent=extent,
               origin='lower',
               aspect='auto')
    plt.ylabel('Time (ns)')
    plt.xlabel('Photon Energy (eV)')
    cbar = plt.colorbar()
    cbar.set_label("Spectrum (GW/sr/eV)")
    plt.show()
    return


def signalImg(signalsArr):
    r"""
    Visualize dante signals as an image.
    
    Parameters
    ----------
    signalsArr: numpy.ndarray
    
    Returns
    -------
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    plt.imshow(signalsArr.transpose(), aspect='auto')
    plt.xlabel('Time')
    plt.ylabel('channel')
    cbar = plt.colorbar()
    cbar.set_label('Signal')
    plt.show()
    return


def individualChPlot(channels, timesFrame, signalsFrame, cut=None):
    r"""
    Plot individual channels on separate plots.
    
    Parameters
    ----------
    
    Returns
    -------
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    if cut:
        timesFrame = timesFrame.iloc[cut:-cut]
        signalsFrame = signalsFrame.iloc[cut:-cut]
    for chNum in channels:
        plt.plot(timesFrame[chNum] * 1e9, signalsFrame[chNum])
        plt.title(f"Ch {chNum}")
        plt.xlabel("Time  (ns)")
        plt.ylabel("Signal (V)")
        plt.show()
    return


def traceGrid(channels,
              timesFrame,
              signalsFrame,
              chStatus=None,
              saveName="",
              cut=None,
              shotNum=""):
    r"""
    Plot individual channels on separate plots arranged into a grid.
    
    Parameters
    ----------
    
    Returns
    -------
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    # cutting leading and trailing edges of signal where noise exists
    if cut:
        timesFrame = timesFrame.iloc[cut:-cut]
        signalsFrame = signalsFrame.iloc[cut:-cut]
        
    # setting up subplots for all 18 channels
    fig, ax = plt.subplots(6, 3, sharex=True, sharey=True, figsize=(15,10))
    fig.subplots_adjust(hspace=0, wspace=0)
    
    # creating array of all possible dante channels
    allCh = np.arange(18) + 1
    # annotating each subplot window with the channel number
    for chNum in allCh:
        col = int((chNum - 1)/6)
        row = (chNum - 1) % 6
        if not chStatus:
            ax[row][col].annotate(f'Ch {chNum}', xy=(0, 0.8), xytext=(0, 0.8))
        else:
            ax[row][col].annotate(f'Ch {chNum} - {chStatus[chNum]}', xy=(0, 0.8), xytext=(0, 0.8))
    # plotting traces for only user specified channels
    for chNum in channels:
        col = int((chNum - 1)/6)
        row = (chNum - 1) % 6
        ax[row][col].plot(timesFrame[chNum] * 1e9,
                          signalsFrame[chNum] / np.max(signalsFrame[chNum]))
        ax[row][col].set_ylim((-0.1, 1.1))
#        ax[row][col].annotate(f'Ch {chNum}', xy=(0, 0.8), xytext=(0, 0.8))
    fig.suptitle(f"Shot {shotNum}")
    fig.text(0.5, 0.04, 'Time (ns, arbitrary offset)', ha='center')
    fig.text(0.04, 0.5, 'Signal (normalized to ch max)', va='center', rotation='vertical')
    
    if saveName:
        # if there is a non-empty name, then try to save
        plt.savefig(saveName)
    else:
        # no name, then just show
        plt.show()
    return


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 26 16:11:09 2019

Fiducia setup script.

@author: Pawel M. Kozlowski
"""

from setuptools import setup


# Read the contents of the README file to include in the long
# description. The long description then becomes part of the pypi.org
# page.
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='fiducia',
      version='0.2.3',
      description='Filtered Diode Unfolder (using) Cubic-spline Algorithm',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/lanl/fiducia',
      author='Pawel M. Kozlowski, et al.',
      author_email='pkozlowski@lanl.gov',
      license='BSD',
      packages=['fiducia'],
      zip_safe=True)